"""Configuration class for function estimators."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict


class FunctionEstimatorConfig(BaseModel):
    """Configuration class for function estimators.

    This class defines the configuration for function estimators, which are
    configured by the `function_estimators` field in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object. That field contains a
    tuple of configuration objects that define which function estimators are
    available during the optimization.

    By default, the final objective and constraint functions and their gradients
    are calculated from the individual realizations by a weighted sum. Function
    estimators are optionally used to modify this calculation.

    The `method` field determines which method will be used to implement the
    calculation of the final function or gradient from the individual
    realizations. To further specify how such a method should function, the
    `options` field can be used to pass a dictionary of key-value pairs. The
    interpretation of these options depends on the chosen method.

    Attributes:
        method:  The function estimator method.
        options: Options to be passed to the estimator.
    """

    method: str = "default/default"
    options: dict[str, Any] = {}

    model_config = ConfigDict(
        extra="forbid",
        str_min_length=1,
        str_strip_whitespace=True,
        validate_default=True,
        frozen=True,
    )



from functools import partial
from typing import TYPE_CHECKING, Any

import numpy as np
import pytest

from ropt.plan import BasicOptimizer
from ropt.plugins.sampler.scipy import _SUPPORTED_METHODS
from ropt.results import GradientResults, Results

if TYPE_CHECKING:
    from numpy.typing import NDArray


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "optimizer": {
            "tolerance": 1e-4,
            "max_functions": 20,
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


@pytest.mark.parametrize("method", sorted(_SUPPORTED_METHODS))
def test_scipy_samplers_unconstrained(
    enopt_config: Any, method: str, evaluator: Any
) -> None:
    enopt_config["samplers"] = [{"method": method}]
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_scipy_indexed_sampler(enopt_config: Any, evaluator: Any) -> None:
    # Removing the second variable will fix its value, since it will not be
    # perturbed and its gradient will always be zero.
    enopt_config["gradient"]["samplers"] = [0, -1, 0]
    enopt_config["variables"]["initial_values"][1] = 0.1

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert pytest.approx(variables[0]) != 0.0
    assert pytest.approx(variables[1]) == 0.1
    assert pytest.approx(variables[2]) != 0.5


@pytest.mark.parametrize("method", sorted(_SUPPORTED_METHODS))
def test_scipy_samplers_shared(enopt_config: Any, method: str, evaluator: Any) -> None:
    enopt_config["realizations"] = {"weights": [1.0, 1.0]}
    enopt_config["samplers"] = [{"method": method}]

    perturbations: dict[str, NDArray[np.float64]] = {}

    def _observer(results: tuple[Results, ...], tag: str) -> None:
        for item in results:
            if isinstance(item, GradientResults) and tag not in perturbations:
                perturbations[tag] = item.evaluations.perturbed_variables

    enopt_config["samplers"][0]["shared"] = False
    variables1 = (
        BasicOptimizer(enopt_config, evaluator())
        .set_results_callback(partial(_observer, tag="result1"))
        .run()
        .variables
    )
    assert variables1 is not None

    enopt_config["samplers"][0]["shared"] = True
    variables2 = (
        BasicOptimizer(enopt_config, evaluator())
        .set_results_callback(partial(_observer, tag="result2"))
        .run()
        .variables
    )
    assert variables2 is not None

    # The perturbations of the two realizations must differ, if not shared:
    assert not np.allclose(
        perturbations["result1"][0, ...], perturbations["result1"][1, ...]
    )

    # The perturbations of the two realizations must be the same, if shared:
    assert np.allclose(
        perturbations["result2"][0, ...], perturbations["result2"][1, ...]
    )

    # The results should be correct, but slightly different:
    assert np.allclose(variables1, [0.0, 0.0, 0.5], atol=0.02)
    assert np.allclose(variables2, [0.0, 0.0, 0.5], atol=0.02)
    assert not np.allclose(variables1, variables2, atol=1e-3)



"""Default values used by the configuration classes."""

from typing import Final

from ropt.enums import BoundaryType, PerturbationType

DEFAULT_SEED: Final = 1
"Default random generator seed."

DEFAULT_NUMBER_OF_PERTURBATIONS: Final = 5
"""Default number of perturbations."""

DEFAULT_PERTURBATION_MAGNITUDE: Final = 0.005
"""Default perturbation magnitude."""

DEFAULT_PERTURBATION_BOUNDARY_TYPE: Final = BoundaryType.MIRROR_BOTH
"""Default perturbation boundary handling type.

See also: [`BoundaryType`][ropt.enums.BoundaryType].
"""

DEFAULT_PERTURBATION_TYPE: Final = PerturbationType.ABSOLUTE
""" Default perturbation type.

See also: [`PerturbationType`][ropt.enums.PerturbationType].
"""



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ropt.enums import ResultAxis

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.transforms import OptModelTransforms


@dataclass(slots=True)
class GradientEvaluations(ResultField):
    """This class contains the results of evaluations for gradient calculation.

    This class stores the variables with the calculated objective and
    constraints gradients. It contains the following information:

    1. The vector of variables at which the functions are evaluated.
    2. A three-dimensional array of perturbed variables, with variable values
       arranged along the third axis. The second axis index indicates the
       perturbation number, whereas the first axis index represents the
       realization number.
    3. The objectives and constraints for each realization and perturbed
       variable vector:  A three-dimensional array, with the objective or
       constraint values arranged along the third axis. The second axis index
       indicates the perturbation number, whereas the first axis index
       represents the realization number.
    4. Optional evaluation IDs that may have been passed from the evaluator,
       identifying each calculated realization and perturbation.

    Attributes:
        variables:                The unperturbed variable vector.
        perturbed_variables:      The variables for each realization and perturbation.
        perturbed_objectives:     The objective functions for each realization and
                                  perturbation.
        perturbed_constraints:    The constraint functions for each realization and
                                  perturbation.
        perturbed_evaluation_ids: Optional id of each evaluated realization.
    """

    variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (ResultAxis.VARIABLE,),
        },
    )
    perturbed_variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxis.REALIZATION,
                ResultAxis.PERTURBATION,
                ResultAxis.VARIABLE,
            ),
        },
    )
    perturbed_objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxis.REALIZATION,
                ResultAxis.PERTURBATION,
                ResultAxis.OBJECTIVE,
            ),
        },
    )
    perturbed_constraints: NDArray[np.float64] | None = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxis.REALIZATION,
                ResultAxis.PERTURBATION,
                ResultAxis.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    perturbed_evaluation_ids: NDArray[np.intc] | None = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxis.REALIZATION,
                ResultAxis.PERTURBATION,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.variables = _immutable_copy(self.variables)
        self.perturbed_variables = _immutable_copy(self.perturbed_variables)
        self.perturbed_objectives = _immutable_copy(self.perturbed_objectives)
        self.perturbed_constraints = _immutable_copy(self.perturbed_constraints)
        self.perturbed_evaluation_ids = _immutable_copy(self.perturbed_evaluation_ids)

    @classmethod
    def create(
        cls,
        variables: NDArray[np.float64],
        perturbed_variables: NDArray[np.float64],
        perturbed_objectives: NDArray[np.float64],
        perturbed_constraints: NDArray[np.float64] | None = None,
        perturbed_evaluation_ids: NDArray[np.intc] | None = None,
    ) -> GradientEvaluations:
        """Create a FunctionEvaluations object with the given information.

        Args:
            config:                   Configuration object.
            variables:                The unperturbed variable vector.
            perturbed_variables:      The unperturbed variable vector.
            perturbed_objectives:     The objective functions for each realization.
            perturbed_constraints:    The constraint functions for each realization.
            perturbed_evaluation_ids: Optional IDs of the objective calculations.

        Returns:
            A new FunctionEvaluations object.
        """
        return GradientEvaluations(
            variables=variables,
            perturbed_variables=perturbed_variables,
            perturbed_objectives=perturbed_objectives,
            perturbed_constraints=perturbed_constraints,
            perturbed_evaluation_ids=perturbed_evaluation_ids,
        )

    def transform_from_optimizer(
        self, transforms: OptModelTransforms
    ) -> GradientEvaluations:
        """Apply transformations from optimizer space.

        Args:
            transforms: The transforms to apply.

        Returns:
            The transformed results.
        """
        return GradientEvaluations(
            variables=(
                self.variables
                if transforms.variables is None
                else transforms.variables.from_optimizer(self.variables)
            ),
            perturbed_variables=(
                self.perturbed_variables
                if transforms.variables is None
                else transforms.variables.from_optimizer(self.perturbed_variables)
            ),
            perturbed_objectives=(
                self.perturbed_objectives
                if transforms.objectives is None
                else transforms.objectives.from_optimizer(self.perturbed_objectives)
            ),
            perturbed_constraints=(
                self.perturbed_constraints
                if (
                    self.perturbed_constraints is None
                    or transforms.nonlinear_constraints is None
                )
                else transforms.nonlinear_constraints.from_optimizer(
                    self.perturbed_constraints
                )
            ),
            perturbed_evaluation_ids=self.perturbed_evaluation_ids,
        )



r"""Code for executing optimization plans.

The [`Plan`][ropt.plan.Plan] class is responsible managing steps and result handlers.

A plan comprises [`PlanStep`][ropt.plugins.plan.base.PlanStep] objects that define
individual actions, as well as
[`ResultHandler`][ropt.plugins.plan.base.ResultHandler] objects that process and
store data generated during execution. Both steps and result handlers are
implemented through a [`plugin`][ropt.plugins.plan] mechanism, making it easy to
extend the range of supported steps and result handlers. The `ropt` library also
includes [default plan handler][ropt.plugins.plan.default.DefaultPlanHandlerPlugin] and
[default plan step][ropt.plugins.plan.default.DefaultPlanStepPlugin]

that provides various steps and result handlers to support a broad range of
optimization workflows.

Most optimization plans require shared state across all steps, such as a
callable for function evaluations or a random number generator. This shared
state is provided through an [`OptimizerContext`][ropt.plan.OptimizerContext]
object, which is supplied when creating the plan.

Setting up and executing a plan object for simple optimization cases can be
complex. The [`BasicOptimizer`][ropt.plan.BasicOptimizer] class simplifies this
process by providing a convenient way to build and execute straightforward plans
involving a single optimization.
"""

from ._basic_optimizer import BasicOptimizer
from ._context import OptimizerContext
from ._events import Event
from ._plan import Plan

__all__ = [
    "BasicOptimizer",
    "Event",
    "OptimizerContext",
    "Plan",
]



"""The main functionality for ensemble-based optimizations.

The [`EnsembleOptimizer`][ropt.optimization.EnsembleOptimizer] class provides
the core functionality for running ensemble-based optimizations. Direct use of
this class is generally discouraged; instead, the [`Plan`][ropt.plan.Plan] or
[`BasicOptimizer`][ropt.plan.BasicOptimizer] classes are recommended for more
flexibility and ease of use.
"""

from ._optimizer import (
    EnsembleOptimizer,
    NestedOptimizerCallback,
    SignalEvaluationCallback,
)

__all__ = [
    "EnsembleOptimizer",
    "NestedOptimizerCallback",
    "SignalEvaluationCallback",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np

from ropt.enums import ResultAxis

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.transforms import OptModelTransforms


@dataclass(slots=True)
class Gradients(ResultField):
    """This class stores the calculated objective and constraint gradients.

    The objective and constraint gradients used by the optimizer are calculated
    from their values evaluated for all realizations, for instance by averaging.
    There may be multiple objectives and constraints. Multiple objectives are
    handled by the optimizer using a weighted sum of the individual gradients,
    stored in the `weighted_objective` field. Multiple constraints are directly
    handled by the optimizer.

    Attributes:
        weighted_objective: The weighted sum of the objective gradients.
        objectives:         The value of each objective gradient.
        constraints:        The value of each constraint gradient.
    """

    weighted_objective: NDArray[np.float64] = field(
        metadata={"__axes__": (ResultAxis.VARIABLE,)},
    )
    objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxis.OBJECTIVE,
                ResultAxis.VARIABLE,
            ),
        },
    )
    constraints: NDArray[np.float64] | None = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxis.NONLINEAR_CONSTRAINT,
                ResultAxis.VARIABLE,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.weighted_objective = _immutable_copy(self.weighted_objective)
        self.objectives = _immutable_copy(self.objectives)
        self.constraints = _immutable_copy(self.constraints)

    @classmethod
    def create(
        cls,
        weighted_objective: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: NDArray[np.float64] | None = None,
    ) -> Gradients:
        """Create a Gradients object with the given information.

        Args:
            config:             Configuration object.
            weighted_objective: The weighted objective.
            objectives:         The objective gradients for each realization.
            constraints:        The constraint gradients for each realization.

        Returns:
            A new Functions object.
        """
        return Gradients(
            weighted_objective=weighted_objective,
            objectives=objectives,
            constraints=constraints,
        )

    def transform_from_optimizer(self, transforms: OptModelTransforms) -> Gradients:
        """Apply transformations from optimizer space.

        Args:
            transforms: The transforms to apply.

        Returns:
            The transformed results.
        """
        objectives = self.objectives
        if transforms.objectives is not None:
            objectives = np.moveaxis(self.objectives, 0, -1)
            objectives = transforms.objectives.from_optimizer(objectives)
            objectives = np.moveaxis(objectives, 0, -1)

        constraints: NDArray[np.float64] | None = self.constraints
        if (
            self.constraints is not None
            and transforms.nonlinear_constraints is not None
        ):
            constraints = np.moveaxis(self.constraints, 0, -1)
            constraints = transforms.nonlinear_constraints.from_optimizer(constraints)
            constraints = np.moveaxis(constraints, 0, -1)

        return Gradients(
            weighted_objective=(
                self.weighted_objective
                if transforms.objectives is None
                else transforms.objectives.weighted_objective_from_optimizer(
                    self.weighted_objective
                )
            ),
            objectives=objectives,
            constraints=constraints,
        )



"""This module implements the SciPy optimization plugin."""

import copy
from functools import partial
from typing import (
    Any,
    Callable,
    ClassVar,
    Final,
)

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import (
    Bounds,
    LinearConstraint,
    NonlinearConstraint,
    differential_evolution,
    minimize,
)

from ropt.config.enopt import EnOptConfig
from ropt.enums import VariableType
from ropt.plugins.optimizer.utils import validate_supported_constraints

from .base import Optimizer, OptimizerCallback, OptimizerPlugin
from .utils import NormalizedConstraints, get_masked_linear_constraints

_SUPPORTED_METHODS: Final[set[str]] = {
    name.lower()
    for name in (
        "Nelder-Mead",
        "Powell",
        "CG",
        "BFGS",
        "Newton-CG",
        "L-BFGS-B",
        "TNC",
        "COBYLA",
        "SLSQP",
        "differential_evolution",
    )
}


# Categorize the methods by the types of constraint they support or require.

_CONSTRAINT_REQUIRES_BOUNDS: Final = {
    "differential_evolution",
}
_CONSTRAINT_SUPPORT_BOUNDS: Final = {
    name.lower()
    for name in [
        "Nelder-Mead",
        "Powell",
        "L-BFGS-B",
        "TNC",
        "SLSQP",
        "differential_evolution",
    ]
}
_CONSTRAINT_SUPPORT_LINEAR_EQ: Final = {
    name.lower() for name in ["SLSQP", "differential_evolution"]
}
_CONSTRAINT_SUPPORT_LINEAR_INEQ: Final = {
    name.lower() for name in ["COBYLA", "SLSQP", "differential_evolution"]
}
_CONSTRAINT_SUPPORT_NONLINEAR_EQ: Final = {
    name.lower() for name in ["SLSQP", "differential_evolution"]
}
_CONSTRAINT_SUPPORT_NONLINEAR_INEQ: Final = {
    name.lower() for name in ["COBYLA", "SLSQP", "differential_evolution"]
}

# These methods do not use a gradient:
_NO_GRADIENT: Final = {
    name.lower()
    for name in ["Nelder-Mead", "Powell", "COBYLA", "differential_evolution"]
}


_ConstraintType = str | Callable[..., float] | Callable[..., NDArray[np.float64]]


class SciPyOptimizer(Optimizer):
    """Plugin class for optimization via SciPy.

    This class implements several optimizers provided by SciPy in the
    [`scipy.optimize`](https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
    package:

    - Nelder-Mead
    - Powell
    - CG
    - BFGS
    - Newton-CG
    - L-BFGS-B
    - TNC
    - COBYLA
    - SLSQP
    - differential_evolution

    The optimizer to use is selected by setting the `method` field in the
    [`optimizer`][ropt.config.enopt.OptimizerConfig] field of
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] to the name of the algorithm.
    Most of these methods support the general options set in the
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object. However, specific
    options that are normally passed as arguments in the SciPy functions can be
    provided via the `options` dictionary in the configuration object. Consult
    the
    [`scipy.optimize`](https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
    manual for details on these options.

    Not all constraints are supported by all optimizers:

    - Bound constraints: Nelder-Mead, L-BFGS-B, SLSQP, TNC,
      differential_evolution
    - Linear constraints: SLSQP, differential_evolution
    - Nonlinear constraints: COBYLA (only inequality), SLSQP,
      differential_evolution

    Info:
        - The Nelder-Mead algorithm only supports bound constraints if SciPy
          version >= 1.7.
        - Some SciPy algorithms that require a Hessian or a Hessian-vector
          product are not supported. These include dogleg, trust-ncg,
          trust-exact, and trust-krylov.
    """

    _supported_constraints: ClassVar[dict[str, set[str]]] = {
        "bounds": _CONSTRAINT_SUPPORT_BOUNDS,
        "linear:eq": _CONSTRAINT_SUPPORT_LINEAR_EQ,
        "linear:ineq": _CONSTRAINT_SUPPORT_LINEAR_INEQ,
        "nonlinear:eq": _CONSTRAINT_SUPPORT_NONLINEAR_EQ,
        "nonlinear:ineq": _CONSTRAINT_SUPPORT_NONLINEAR_INEQ,
    }
    _required_constraints: ClassVar[dict[str, set[str]]] = {
        "bounds": _CONSTRAINT_REQUIRES_BOUNDS,
    }

    def __init__(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> None:
        """Initialize the optimizer implemented by the SciPy plugin.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._optimizer_callback = optimizer_callback
        self._config = config
        _, _, self._method = self._config.optimizer.method.lower().rpartition("/")
        if self._method == "default":
            self._method = "slsqp"
        if self._method not in _SUPPORTED_METHODS:
            msg = f"SciPy optimizer algorithm {self._method} is not supported"
            raise NotImplementedError(msg)
        validate_supported_constraints(
            self._config,
            self._method,
            self._supported_constraints,
            self._required_constraints,
        )
        self._bounds = self._initialize_bounds()
        self._constraints = self._initialize_constraints()
        self._options = self._parse_options()
        self._parallel = (
            self._config.optimizer.parallel and self._method == "differential_evolution"
        )

        self._cached_variables: NDArray[np.float64] | None = None
        self._cached_function: NDArray[np.float64] | None = None
        self._cached_gradient: NDArray[np.float64] | None = None

    def start(self, initial_values: NDArray[np.float64]) -> None:
        """Start the optimization.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._cached_variables = None
        self._cached_function = None
        self._cached_gradient = None

        if self._config.variables.mask is not None:
            initial_values = initial_values[..., self._config.variables.mask]

        if self._method == "differential_evolution":
            if self._parallel:
                self._options["updating"] = "deferred"
                self._options["workers"] = 1
            differential_evolution(
                func=self._function,
                x0=initial_values,
                bounds=self._bounds,
                constraints=self._constraints,
                polish=False,
                vectorized=self._parallel,
                **self._options,
            )
        else:
            minimize(
                fun=self._function,
                x0=initial_values,
                tol=self._config.optimizer.tolerance,
                method=self._method,
                bounds=self._bounds,
                jac=(False if self._method in _NO_GRADIENT else self._gradient),
                constraints=self._constraints,
                options=self._options if self._options else None,
            )

    @property
    def allow_nan(self) -> bool:
        """Whether NaN is allowed.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._method == "differential_evolution"

    @property
    def is_parallel(self) -> bool:
        """Whether the current run is parallel.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._parallel

    def _initialize_bounds(self) -> Bounds | None:
        if (
            np.isfinite(self._config.variables.lower_bounds).any()
            or np.isfinite(self._config.variables.upper_bounds).any()
        ):
            lower_bounds = self._config.variables.lower_bounds
            upper_bounds = self._config.variables.upper_bounds
            if self._config.variables.mask is not None:
                lower_bounds = lower_bounds[self._config.variables.mask]
                upper_bounds = upper_bounds[self._config.variables.mask]
            return Bounds(lower_bounds, upper_bounds)
        return None

    def _initialize_constraints(
        self,
    ) -> list[dict[str, _ConstraintType] | NonlinearConstraint | LinearConstraint]:
        self._normalized_constraints = None

        lin_coef, lin_lower, lin_upper = None, None, None
        if self._config.linear_constraints is not None:
            lin_coef, lin_lower, lin_upper = get_masked_linear_constraints(self._config)

        if self._method == "differential_evolution":
            return self._initialize_constraints_object(lin_coef, lin_lower, lin_upper)

        lower_bounds = []
        upper_bounds = []
        if self._config.nonlinear_constraints is not None:
            lower_bounds.append(self._config.nonlinear_constraints.lower_bounds)
            upper_bounds.append(self._config.nonlinear_constraints.upper_bounds)
        if lin_lower is not None and lin_upper is not None:
            lower_bounds.append(lin_lower)
            upper_bounds.append(lin_upper)
        if lower_bounds:
            self._normalized_constraints = NormalizedConstraints(
                np.concatenate(lower_bounds), np.concatenate(upper_bounds)
            )
            return self._initialize_constraints_dict(lin_coef)

        return []

    def _fun(
        self,
        variables: NDArray[np.float64],
        index: int,
        lin_coef: NDArray[np.float64] | None,
    ) -> NDArray[np.float64]:
        assert self._normalized_constraints is not None
        if self._normalized_constraints.constraints is None:
            constraints = []
            if self._config.nonlinear_constraints is not None:
                constraints.append(self._constraint_functions(variables).transpose())
            if lin_coef is not None:
                constraints.append(np.matmul(lin_coef, variables))
            self._normalized_constraints.set_constraints(
                np.concatenate(constraints, axis=0)
            )
        assert self._normalized_constraints.constraints is not None
        return self._normalized_constraints.constraints[index, :]

    def _jac(
        self,
        variables: NDArray[np.float64],
        index: int,
        lin_coef: NDArray[np.float64] | None,
    ) -> NDArray[np.float64]:
        assert self._normalized_constraints is not None
        if self._normalized_constraints.gradients is None:
            gradients = []
            if self._config.nonlinear_constraints is not None:
                gradients.append(self._constraint_gradients(variables))
            if lin_coef is not None:
                gradients.append(lin_coef)
            self._normalized_constraints.set_gradients(
                np.concatenate(gradients, axis=0)
            )
        assert self._normalized_constraints.gradients is not None
        return self._normalized_constraints.gradients[index, :]

    def _initialize_constraints_dict(
        self,
        lin_coef: NDArray[np.float64] | None,
    ) -> list[dict[str, _ConstraintType]]:
        if self._normalized_constraints is None:
            return []

        def _constraint_entry(type_: str, index: int) -> dict[str, _ConstraintType]:
            fun = partial(self._fun, index=index, lin_coef=lin_coef)
            if self._method == "cobyla":
                return {"type": type_, "fun": fun}
            jac = partial(self._jac, index=index, lin_coef=lin_coef)
            return {"type": type_, "fun": fun, "jac": jac}

        return [
            _constraint_entry("eq" if is_eq else "ineq", inx)
            for inx, is_eq in enumerate(self._normalized_constraints.is_eq)
        ]

    def _initialize_constraints_object(
        self,
        lin_coef: NDArray[np.float64] | None,
        lin_lower: NDArray[np.float64] | None,
        lin_upper: NDArray[np.float64] | None,
    ) -> list[LinearConstraint | NonlinearConstraint]:
        constraints = []
        if self._config.linear_constraints is not None:
            constraints.append(LinearConstraint(lin_coef, lin_lower, lin_upper))

        if self._config.nonlinear_constraints is not None:
            lower_bounds = self._config.nonlinear_constraints.lower_bounds
            upper_bounds = self._config.nonlinear_constraints.upper_bounds

            def _fun(variables: NDArray[np.float64]) -> NDArray[np.float64]:
                functions = self._constraint_functions(variables)
                return functions.transpose()

            def _jac(variables: NDArray[np.float64]) -> NDArray[np.float64]:
                return self._constraint_gradients(variables)

            constraints.append(
                NonlinearConstraint(
                    fun=_fun, jac=_jac, lb=lower_bounds, ub=upper_bounds
                ),
            )
        return constraints

    def _function(self, variables: NDArray[np.float64]) -> NDArray[np.float64]:
        if variables.ndim > 1 and variables.size == 0:
            return np.array([])
        functions, _ = self._get_function_or_gradient(
            variables, get_function=True, get_gradient=False
        )
        assert functions is not None
        if variables.ndim > 1:
            return functions[:, 0]
        return np.array(functions[0])

    def _gradient(self, variables: NDArray[np.float64]) -> NDArray[np.float64]:
        _, gradients = self._get_function_or_gradient(
            variables, get_function=False, get_gradient=True
        )
        assert gradients is not None
        return gradients[0, :]

    def _constraint_functions(
        self, variables: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        if variables.ndim > 1 and variables.size == 0:
            return np.array([])
        functions, _ = self._get_function_or_gradient(
            variables, get_function=True, get_gradient=False
        )
        assert functions is not None
        if variables.ndim > 1:
            return functions[:, 1:]
        return np.array(functions[1:])

    def _constraint_gradients(
        self, variables: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        _, gradients = self._get_function_or_gradient(
            variables, get_function=False, get_gradient=True
        )
        assert gradients is not None
        return gradients[1:, :]

    def _get_function_or_gradient(
        self, variables: NDArray[np.float64], *, get_function: bool, get_gradient: bool
    ) -> tuple[NDArray[np.float64] | None, NDArray[np.float64] | None]:
        if self._parallel and variables.ndim > 1:
            variables = variables.T

        if self._method in _NO_GRADIENT:
            get_gradient = False

        if (
            self._cached_variables is None
            or variables.shape != self._cached_variables.shape
            or not np.allclose(variables, self._cached_variables)
        ):
            self._cached_variables = None
            self._cached_function = None
            self._cached_gradient = None
            if self._normalized_constraints is not None:
                self._normalized_constraints.reset()

        function = self._cached_function if get_function else None
        gradient = self._cached_gradient if get_gradient else None

        compute_functions = get_function and function is None
        compute_gradients = get_gradient and gradient is None

        if compute_functions or compute_gradients:
            self._cached_variables = variables.copy()
            compute_functions = compute_functions or self._config.optimizer.speculative
            compute_gradients = compute_gradients or self._config.optimizer.speculative
            new_function, new_gradient = self._compute_functions_and_gradients(
                variables,
                compute_functions=compute_functions,
                compute_gradients=compute_gradients,
            )
            if compute_functions:
                assert new_function is not None
                self._cached_function = new_function.copy()
                if get_function:
                    function = new_function
            if compute_gradients:
                assert new_gradient is not None
                self._cached_gradient = new_gradient.copy()
                if get_gradient:
                    gradient = new_gradient

        return function, gradient

    def _compute_functions_and_gradients(
        self,
        variables: NDArray[np.float64],
        *,
        compute_functions: bool,
        compute_gradients: bool,
    ) -> tuple[NDArray[np.float64] | None, NDArray[np.float64] | None]:
        new_function = None
        new_gradient = None
        if (
            compute_functions
            and compute_gradients
            and self._config.optimizer.split_evaluations
        ):
            new_function, _ = self._optimizer_callback(
                variables,
                return_functions=True,
                return_gradients=False,
            )
            _, new_gradient = self._optimizer_callback(
                variables,
                return_functions=False,
                return_gradients=True,
            )
        else:
            new_function, new_gradient = self._optimizer_callback(
                variables,
                return_functions=compute_functions,
                return_gradients=compute_gradients,
            )
        if self.allow_nan:
            new_function = np.where(np.isnan(new_function), np.inf, new_function)
        return new_function, new_gradient

    def _parse_options(self) -> dict[str, Any]:
        if not isinstance(self._config.optimizer.options, dict):
            return {}
        options = copy.deepcopy(self._config.optimizer.options)
        # The maximum number of iterations is passed as an option to ropt.
        # Setting maxiter directly as an entry in the options dict will also
        # work, but iterations will override it.
        iterations = self._config.optimizer.max_iterations
        if iterations is not None:
            if self._method == "tnc":
                options["maxfun"] = iterations
            else:
                options["maxiter"] = iterations
        # We switch on display if there is an output folder.
        if self._config.optimizer.output_dir is not None:
            options["disp"] = True

        if (
            self._method == "differential_evolution"
            and self._config.variables.types is not None
            and "integrality" not in options
        ):
            options["integrality"] = (
                self._config.variables.types == VariableType.INTEGER
            )

        return options


class SciPyOptimizerPlugin(OptimizerPlugin):
    """The SciPY optimizer plugin class."""

    def create(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> SciPyOptimizer:
        """Initialize the optimizer plugin.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        return SciPyOptimizer(config, optimizer_callback)

    def is_supported(self, method: str) -> bool:
        """Check if a method is supported.

        See the [ropt.plugins.base.Plugin][] abstract base class.

        # noqa
        """
        return method.lower() in (_SUPPORTED_METHODS | {"default"})



from typing import Any, cast

import numpy as np
import pytest
from numpy.typing import NDArray

from ropt.plan import BasicOptimizer

pytestmark = [pytest.mark.slow]


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "optimizer": {
            "method": "differential_evolution",
            "tolerance": 1e-4,
            "max_iterations": 20,
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_scipy_required_constraints_bounds_de(
    enopt_config: Any, evaluator: Any
) -> None:
    with pytest.raises(NotImplementedError, match="requires bound constraints"):
        BasicOptimizer(enopt_config, evaluator()).run()


def test_scipy_bound_constraints_de(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["lower_bounds"] = [0.15, 0.0, 0.0]
    enopt_config["variables"]["upper_bounds"] = [0.5, 0.5, 0.2]
    enopt_config["variables"]["initial_values"][0] = 0.2
    enopt_config["optimizer"]["options"] = {"seed": 1}

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.15, 0.0, 0.2], atol=0.02)


def test_scipy_bound_constraints_differential_evolution_de(
    enopt_config: Any, evaluator: Any, test_functions: Any
) -> None:
    enopt_config["variables"]["lower_bounds"] = [0.15, 0.0, 0.0]
    enopt_config["variables"]["upper_bounds"] = [0.5, 0.5, 0.2]
    enopt_config["variables"]["initial_values"][0] = 0.2
    enopt_config["optimizer"]["options"] = {"seed": 1}

    enopt_config["realizations"] = {"realization_min_success": 0}
    variables1 = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables1 is not None
    assert np.allclose(variables1, [0.15, 0.0, 0.2], atol=0.025)

    counter = 0

    def _add_nan(x: Any, c: Any) -> Any:
        nonlocal counter
        counter += 1
        if counter == 2:
            counter = 0
            return np.nan
        return test_functions[0](x, c)

    variables2 = (
        BasicOptimizer(enopt_config, evaluator((_add_nan, test_functions[1])))
        .run()
        .variables
    )
    assert variables2 is not None
    assert np.allclose(variables2, [0.15, 0.0, 0.2], atol=0.025)
    assert not np.all(variables1 == variables2)


def test_scipy_eq_linear_constraints_de(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1], [0, 1, 1]],
        "lower_bounds": [1.0, 0.75],
        "upper_bounds": [1.0, 0.75],
    }

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.25, 0.0, 0.75], atol=0.02)


def test_scipy_ge_linear_constraints_de(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["linear_constraints"] = {
        "coefficients": [[-1, 0, -1]],
        "lower_bounds": -0.4,
        "upper_bounds": np.inf,
    }
    enopt_config["optimizer"]["options"] = {"seed": 1}

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.05, 0.0, 0.45], atol=0.02)


def test_scipy_le_linear_constraints_de(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1]],
        "lower_bounds": -np.inf,
        "upper_bounds": 0.4,
    }
    enopt_config["optimizer"]["options"] = {"seed": 1}

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.05, 0.0, 0.45], atol=0.02)


def test_scipy_le_ge_linear_constraints_de(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1], [-1, 0, -1]],
        "lower_bounds": [-np.inf, -0.4],
        "upper_bounds": [0.4, np.inf],
    }
    enopt_config["optimizer"]["options"] = {"seed": 1}

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.05, 0.0, 0.45], atol=0.02)


def test_scipy_le_ge_linear_constraints_two_sided_de(
    enopt_config: Any, evaluator: Any
) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1], [1, 0, 1]],
        "lower_bounds": [-np.inf, 0.0],
        "upper_bounds": [0.3, np.inf],
    }
    enopt_config["optimizer"]["options"] = {"seed": 1}

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.1, 0.0, 0.4], atol=0.02)

    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1]],
        "lower_bounds": [0.0],
        "upper_bounds": [0.3],
    }

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.1, 0.0, 0.4], atol=0.02)


def test_scipy_eq_nonlinear_constraints_de(
    enopt_config: Any, evaluator: Any, test_functions: Any
) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": 1.0,
        "upper_bounds": 1.0,
    }

    test_functions = (
        *test_functions,
        lambda variables, _: cast(NDArray[np.float64], variables[0] + variables[2]),
    )

    variables = BasicOptimizer(enopt_config, evaluator(test_functions)).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.25, 0.0, 0.75], atol=0.02)


@pytest.mark.parametrize(
    ("lower_bounds", "upper_bounds"), [(-np.inf, 0.4), (-0.4, np.inf)]
)
def test_scipy_ineq_nonlinear_constraints_de(
    enopt_config: Any,
    lower_bounds: Any,
    upper_bounds: Any,
    evaluator: Any,
    test_functions: Any,
) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]

    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": lower_bounds,
        "upper_bounds": upper_bounds,
    }
    enopt_config["optimizer"]["options"] = {"seed": 1}

    weight = 1.0 if upper_bounds == 0.4 else -1.0
    test_functions = (
        *test_functions,
        lambda variables, _: cast(
            NDArray[np.float64], weight * variables[0] + weight * variables[2]
        ),
    )

    variables = BasicOptimizer(enopt_config, evaluator(test_functions)).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.05, 0.0, 0.45], atol=0.02)


def test_scipy_ineq_nonlinear_constraints_two_sided_de(
    enopt_config: Any,
    evaluator: Any,
    test_functions: Any,
) -> None:
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]
    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": [0.0],
        "upper_bounds": [0.3],
    }
    enopt_config["optimizer"]["options"] = {"seed": 1}
    test_functions = (
        *test_functions,
        lambda variables, _: cast(NDArray[np.float64], variables[0] + variables[2]),
    )

    variables = BasicOptimizer(enopt_config, evaluator(test_functions)).run().variables
    assert variables is not None
    assert np.allclose(variables, [-0.1, 0.0, 0.4], atol=0.02)


def test_scipy_le_ge_nonlinear_constraints_de(
    enopt_config: Any,
    evaluator: Any,
    test_functions: Any,
) -> None:
    # These constraints together force the first two variables to be zero,
    # while the last one is free to fit the function.
    enopt_config["variables"]["lower_bounds"] = [-1.0, -1.0, -1.0]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.0, 1.0]
    enopt_config["variables"]["lower_bounds"] = 0.0

    enopt_config["nonlinear_constraints"] = {
        "lower_bounds": [0.4, 0.0],
        "upper_bounds": [np.inf, 0.0],
    }

    test_functions = (
        *test_functions,
        lambda variables, _: cast(NDArray[np.float64], variables[0] + variables[2]),
    )
    test_functions = (
        *test_functions,
        lambda variables, _: cast(NDArray[np.float64], variables[0] - variables[1]),
    )

    variables = BasicOptimizer(enopt_config, evaluator(test_functions)).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)



"""Configuration class for realization filters."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict


class RealizationFilterConfig(BaseModel):
    """The configuration class for realization filters.

    This class defines the configuration for realization filters, which are
    configured by the `realization_filters` field in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object. This field contains a
    tuple of configuration objects that define which realization filters are
    available during optimization.

    By default, the final objective and constraint functions and their gradients
    are calculated as a weighted function from all realizations. Realization
    filters are optionally used to change the weights of the individual
    realizations. For instance, this can be used to determine which subset of
    realizations should be used in calculating the final objective and
    constraint functions and their gradients by setting some weights to zero.

    The `method` field determines which method will be used to adjust the
    weights of the individual realizations. To further specify how such a method
    should function, the `options` field can be used to pass a dictionary of
    key-value pairs. The interpretation of these options depends on the| chosen
    method.

    Attributes:
        method:  The realization filter method.
        options: Options to be passed to the filter.
    """

    method: str
    options: dict[str, Any] = {}

    model_config = ConfigDict(
        extra="forbid",
        str_min_length=1,
        str_strip_whitespace=True,
        validate_default=True,
        frozen=True,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ropt.enums import ResultAxis

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.transforms import OptModelTransforms


@dataclass(slots=True)
class FunctionEvaluations(ResultField):
    """This class contains the results of evaluations for function calculation.

    This class stores the variables with the calculated objective and constraint
    functions. It contains the following information:

    1. The vector of variables at which the functions are evaluated.
    2. The calculated objectives and constraints for each realization: A
       two-dimensional array, with the objective or constraint values arranged
       along the second axis. The first axis index indicates the realization
       number.
    3. Optional evaluation IDs that may have been passed from the evaluator,
       identifying each calculated realization.

    Attributes:
        variables:      The unperturbed variable vector.
        objectives:     The objective functions for each realization.
        constraints:    The constraint functions for each realization.
        evaluation_ids: Optional id of each evaluated realization.
    """

    variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (ResultAxis.VARIABLE,),
        },
    )
    objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxis.REALIZATION,
                ResultAxis.OBJECTIVE,
            ),
        },
    )
    constraints: NDArray[np.float64] | None = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxis.REALIZATION,
                ResultAxis.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    evaluation_ids: NDArray[np.intc] | None = field(
        default=None,
        metadata={
            "__axes__": (ResultAxis.REALIZATION,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.variables = _immutable_copy(self.variables)
        self.objectives = _immutable_copy(self.objectives)
        self.constraints = _immutable_copy(self.constraints)
        self.evaluation_ids = _immutable_copy(self.evaluation_ids)

    @classmethod
    def create(
        cls,
        variables: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: NDArray[np.float64] | None = None,
        evaluation_ids: NDArray[np.intc] | None = None,
    ) -> FunctionEvaluations:
        """Create a FunctionEvaluations object with the given information.

        Args:
            config:                 Configuration object.
            variables:              The unperturbed variable vector.
            objectives:             The objective functions for each realization.
            constraints:            The constraint functions for each realization.
            evaluation_ids:         Optional IDs of the objective calculations.

        Returns:
            A new FunctionEvaluations object.
        """
        return FunctionEvaluations(
            variables=variables,
            objectives=objectives,
            constraints=constraints,
            evaluation_ids=evaluation_ids,
        )

    def transform_from_optimizer(
        self, transforms: OptModelTransforms
    ) -> FunctionEvaluations:
        """Apply transformations from optimizer space.

        Args:
            transforms: The transforms to apply.

        Returns:
            The transformed results.
        """
        return FunctionEvaluations(
            variables=(
                self.variables
                if transforms.variables is None
                else transforms.variables.from_optimizer(self.variables)
            ),
            objectives=(
                self.objectives
                if transforms.objectives is None
                else transforms.objectives.from_optimizer(self.objectives)
            ),
            constraints=(
                self.constraints
                if self.constraints is None or transforms.nonlinear_constraints is None
                else transforms.nonlinear_constraints.from_optimizer(self.constraints)
            ),
            evaluation_ids=self.evaluation_ids,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from numpy.random import default_rng

from ropt.results import (
    ConstraintInfo,
    FunctionEvaluations,
    FunctionResults,
    Functions,
    GradientEvaluations,
    GradientResults,
    Gradients,
    Realizations,
    Results,
)

from ._evaluator_results import (
    _FunctionEvaluatorResults,
    _get_active_realizations,
    _get_function_and_gradient_results,
    _get_function_results,
    _get_gradient_results,
)
from ._function import _calculate_estimated_constraints, _calculate_estimated_objectives
from ._gradient import (
    _calculate_estimated_constraint_gradients,
    _calculate_estimated_objective_gradients,
    _perturb_variables,
)
from ._utils import _get_failed_realizations

if TYPE_CHECKING:
    from numpy.random import Generator
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig
    from ropt.evaluator import Evaluator
    from ropt.plugins import PluginManager
    from ropt.plugins.function_estimator.base import FunctionEstimator
    from ropt.plugins.realization_filter.base import RealizationFilter
    from ropt.plugins.sampler.base import Sampler
    from ropt.transforms import OptModelTransforms


class EnsembleEvaluator:
    """A class for constructing functions and gradients from an ensemble of functions.

    This class implements the calculation of functions and gradients from an
    ensemble of functions, based on the settings defined in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] configuration object. It uses
    an [`Evaluator`][ropt.evaluator.Evaluator] callable to evaluate the individual
    functions.
    """

    def __init__(
        self,
        config: EnOptConfig,
        transforms: OptModelTransforms | None,
        evaluator: Evaluator,
        plan_id: tuple[int, ...],
        plugin_manager: PluginManager,
    ) -> None:
        """Initialize the ensemble evaluator.

        Args:
            config:         The configuration object.
            transforms:     Optional transforms object.
            evaluator:      The callable for evaluation individual functions.
            plan_id:        A tuple identifying the plan running this evaluator.
            plugin_manager: A plugin manager to load required plugins.
        """
        self._config = config
        self._transforms = transforms
        self._evaluator = evaluator
        self._plan_id = plan_id
        self._realization_filters = self._init_realization_filters(plugin_manager)
        self._function_estimators = self._init_function_estimators(plugin_manager)
        rng = default_rng(config.gradient.seed)
        self._samplers = self._init_samplers(rng, plugin_manager)
        self._cache_for_gradient: FunctionResults | None = None

    def calculate(
        self,
        variables: NDArray[np.float64],
        *,
        compute_functions: bool,
        compute_gradients: bool,
    ) -> tuple[Results, ...]:
        """Evaluate the given variable vectors.

        The `variables` argument may be a single vector of variables or a set
        of variable vectors represented as row-vectors in a matrix. The
        `compute_functions` and `compute_gradients` flags determine which
        results are returned: functions, gradients, or both.

        Args:
            variables:         The variable vectors to evaluate.
            compute_functions: Whether to calculate functions.
            compute_gradients: Whether to calculate gradients.

        Returns:
            The results for function evaluations and/or gradient evaluations.
        """
        assert compute_functions or compute_gradients

        # Only functions:
        if compute_functions and not compute_gradients:
            return self._calculate_functions(variables)

        # Parallel evaluation is not supported for gradient-based optimizers:
        assert variables.ndim == 1

        # Only a gradient, and there is a cached function value:
        if (
            compute_gradients
            and not compute_functions
            and self._cache_for_gradient is not None
            and np.allclose(
                self._cache_for_gradient.evaluations.variables,
                variables,
                rtol=0.0,
                atol=1e-15,
            )
        ):
            # This assumes that the cached function value is calculated with the
            # same parameters as the gradient is to be calculated. This is true
            # because each optimization step is a single optimizer run and uses
            # its own copy of an `EnsembleFunctionEvaluation` object.
            return self._calculate_gradients(variables, self._config.variables.mask)

        # A function + gradient, or a gradient without cached function:
        self._cache_for_gradient = None
        return self._calculate_both(variables, self._config.variables.mask)

    def _calculate_functions(
        self, variables: NDArray[np.float64]
    ) -> tuple[FunctionResults, ...]:
        if variables.ndim == 1:
            variables = variables[np.newaxis, :]
        active_objectives, active_constraints = _get_active_realizations(self._config)
        function_results = tuple(
            self._calculate_one_set_of_functions(f_eval_results, variables[idx, :])
            for idx, f_eval_results in _get_function_results(
                self._config,
                self._transforms,
                self._evaluator,
                variables,
                active_objectives,
                active_constraints,
            )
        )

        # Gradient-based methods are currently not parallelized, and there is
        # only one function result. That function may be needed in a gradient
        # calculation, so it is cached here:
        self._cache_for_gradient = function_results[0]

        return function_results

    def _calculate_one_set_of_functions(
        self, f_eval_results: _FunctionEvaluatorResults, variables: NDArray[np.float64]
    ) -> FunctionResults:
        (
            objective_weights,
            constraint_weights,
        ) = self._calculate_filtered_realization_weights(f_eval_results)

        assert self._config.gradient.perturbation_min_success is not None
        failed_realizations = _get_failed_realizations(
            f_eval_results.objectives,
            None,
            self._config.gradient.perturbation_min_success,
        )

        assert self._config.realizations.realization_min_success is not None
        if (
            np.count_nonzero(~failed_realizations)
            >= self._config.realizations.realization_min_success
        ):
            functions = self._compute_functions(
                f_eval_results.objectives,
                f_eval_results.constraints,
                objective_weights,
                constraint_weights,
                failed_realizations,
            )
        else:
            functions = None

        evaluations = FunctionEvaluations.create(
            variables=variables,
            objectives=f_eval_results.objectives,
            constraints=f_eval_results.constraints,
            evaluation_ids=f_eval_results.evaluation_ids,
        )

        return FunctionResults(
            plan_id=self._plan_id,
            batch_id=f_eval_results.batch_id,
            metadata={},
            evaluations=evaluations,
            realizations=Realizations(
                failed_realizations=failed_realizations,
                objective_weights=objective_weights,
                constraint_weights=constraint_weights,
            ),
            functions=functions,
            constraint_info=ConstraintInfo.create(
                self._config,
                evaluations.variables,
                functions.constraints if functions is not None else None,
            ),
        )

    def _calculate_gradients(
        self,
        variables: NDArray[np.float64],
        mask: NDArray[np.bool_] | None,
    ) -> tuple[GradientResults]:
        perturbed_variables = _perturb_variables(
            self._config, variables, self._samplers
        )

        # No functions are computed in this case, instead they must have been
        # computed in a previous run, with the results stored in the
        # cached_result argument. In that case, we can skip any realizations
        # that have a weight equal to zero, we we pull those from the cached
        # results.
        assert self._cache_for_gradient is not None
        assert self._cache_for_gradient.realizations is not None
        objective_weights = self._cache_for_gradient.realizations.objective_weights
        constraint_weights = self._cache_for_gradient.realizations.constraint_weights

        active_objectives, active_constraints = _get_active_realizations(
            self._config,
            objective_weights=objective_weights,
            constraint_weights=constraint_weights,
        )
        g_eval_results = _get_gradient_results(
            self._config,
            self._transforms,
            self._evaluator,
            perturbed_variables,
            active_objectives,
            active_constraints,
        )

        assert self._config.gradient.perturbation_min_success is not None
        failed_realizations = _get_failed_realizations(
            self._cache_for_gradient.evaluations.objectives,
            g_eval_results.perturbed_objectives,
            self._config.gradient.perturbation_min_success,
        )
        assert self._config.realizations.realization_min_success is not None
        if (
            np.count_nonzero(~failed_realizations)
            >= self._config.realizations.realization_min_success
        ):
            gradients = self._compute_gradients(
                variables,
                mask,
                perturbed_variables,
                self._cache_for_gradient.evaluations.objectives,
                self._cache_for_gradient.evaluations.constraints,
                g_eval_results.perturbed_objectives,
                g_eval_results.perturbed_constraints,
                objective_weights,
                constraint_weights,
                failed_realizations,
            )
        else:
            gradients = None

        assert g_eval_results.perturbed_objectives is not None
        return (
            GradientResults(
                plan_id=self._plan_id,
                batch_id=g_eval_results.batch_id,
                metadata={},
                evaluations=GradientEvaluations.create(
                    variables=variables,
                    perturbed_variables=perturbed_variables,
                    perturbed_objectives=g_eval_results.perturbed_objectives,
                    perturbed_constraints=g_eval_results.perturbed_constraints,
                    perturbed_evaluation_ids=g_eval_results.perturbed_evaluation_ids,
                ),
                realizations=Realizations(
                    failed_realizations=failed_realizations,
                    objective_weights=objective_weights,
                    constraint_weights=constraint_weights,
                ),
                gradients=gradients,
            ),
        )

    def _calculate_both(
        self,
        variables: NDArray[np.float64],
        mask: NDArray[np.bool_] | None,
    ) -> tuple[FunctionResults, GradientResults]:
        perturbed_variables = _perturb_variables(
            self._config, variables, self._samplers
        )
        active_objectives, active_constraints = _get_active_realizations(self._config)
        f_eval_results, g_eval_results = _get_function_and_gradient_results(
            self._config,
            self._transforms,
            self._evaluator,
            variables,
            perturbed_variables,
            active_objectives,
            active_constraints,
        )

        evaluations = FunctionEvaluations.create(
            variables=variables,
            objectives=f_eval_results.objectives,
            constraints=f_eval_results.constraints,
            evaluation_ids=f_eval_results.evaluation_ids,
        )

        (
            objective_weights,
            constraint_weights,
        ) = self._calculate_filtered_realization_weights(
            f_eval_results,
        )

        assert self._config.gradient.perturbation_min_success is not None
        failed_realizations = _get_failed_realizations(
            f_eval_results.objectives,
            None,
            self._config.gradient.perturbation_min_success,
        )
        assert self._config.realizations.realization_min_success is not None
        if (
            np.count_nonzero(~failed_realizations)
            >= self._config.realizations.realization_min_success
        ):
            functions = self._compute_functions(
                f_eval_results.objectives,
                f_eval_results.constraints,
                objective_weights,
                constraint_weights,
                failed_realizations,
            )
        else:
            functions = None

        function_results = FunctionResults(
            plan_id=self._plan_id,
            batch_id=f_eval_results.batch_id,
            metadata={},
            evaluations=evaluations,
            realizations=Realizations(
                failed_realizations=failed_realizations,
                objective_weights=objective_weights,
                constraint_weights=constraint_weights,
            ),
            functions=functions,
            constraint_info=ConstraintInfo.create(
                self._config,
                evaluations.variables,
                functions.constraints if functions is not None else None,
            ),
        )

        assert self._config.gradient.perturbation_min_success is not None
        failed_realizations = _get_failed_realizations(
            f_eval_results.objectives,
            g_eval_results.perturbed_objectives,
            self._config.gradient.perturbation_min_success,
        )
        assert self._config.realizations.realization_min_success is not None
        if (
            np.count_nonzero(~failed_realizations)
            >= self._config.realizations.realization_min_success
        ):
            gradients = self._compute_gradients(
                variables,
                mask,
                perturbed_variables,
                f_eval_results.objectives,
                f_eval_results.constraints,
                g_eval_results.perturbed_objectives,
                g_eval_results.perturbed_constraints,
                objective_weights,
                constraint_weights,
                failed_realizations,
            )
        else:
            gradients = None

        gradient_results = GradientResults(
            plan_id=self._plan_id,
            batch_id=g_eval_results.batch_id,
            metadata={},
            evaluations=GradientEvaluations.create(
                variables=variables,
                perturbed_variables=perturbed_variables,
                perturbed_objectives=g_eval_results.perturbed_objectives,
                perturbed_constraints=g_eval_results.perturbed_constraints,
                perturbed_evaluation_ids=g_eval_results.perturbed_evaluation_ids,
            ),
            realizations=Realizations(
                failed_realizations=failed_realizations,
                objective_weights=objective_weights,
                constraint_weights=constraint_weights,
            ),
            gradients=gradients,
        )

        return function_results, gradient_results

    def _compute_functions(
        self,
        objectives: NDArray[np.float64],
        constraints: NDArray[np.float64] | None,
        objective_weights: NDArray[np.float64] | None,
        constraint_weights: NDArray[np.float64] | None,
        failed_realizations: NDArray[np.bool_],
    ) -> Functions:
        # Individual objective and constraint functions are calculated from the
        # realizations using one or more function estimators:
        if np.all(failed_realizations):
            objectives = np.empty(
                self._config.objectives.weights.size, dtype=np.float64
            )
            objectives.fill(np.nan)
            if constraints is not None:
                assert self._config.nonlinear_constraints is not None
                constraints = np.empty(
                    self._config.nonlinear_constraints.lower_bounds.shape,
                    dtype=np.float64,
                )
                constraints.fill(np.nan)
            weighted_objective = np.array(np.nan)
        else:
            objectives = _calculate_estimated_objectives(
                self._config,
                self._function_estimators,
                objectives,
                objective_weights,
                failed_realizations,
            )
            constraints = _calculate_estimated_constraints(
                self._config,
                self._function_estimators,
                constraints,
                constraint_weights,
                failed_realizations,
            )

            weighted_objective = np.array(
                (self._config.objectives.weights * objectives).sum()
            )

        return Functions.create(
            weighted_objective=weighted_objective,
            objectives=objectives,
            constraints=constraints,
        )

    def _compute_gradients(  # noqa: PLR0913
        self,
        variables: NDArray[np.float64],
        mask: NDArray[np.bool_] | None,
        perturbed_variables: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: NDArray[np.float64] | None,
        perturbed_objectives: NDArray[np.float64],
        perturbed_constraints: NDArray[np.float64] | None,
        objective_weights: NDArray[np.float64] | None,
        constraint_weights: NDArray[np.float64] | None,
        failed_realizations: NDArray[np.bool_],
    ) -> Gradients:
        if mask is not None:
            variables = variables[mask]
        variables = np.repeat(
            variables[np.newaxis, ...], self._config.realizations.weights.size, axis=0
        )

        if mask is not None:
            perturbed_variables = perturbed_variables[..., mask]

        assert perturbed_objectives is not None
        objective_gradients = _calculate_estimated_objective_gradients(
            self._config,
            self._function_estimators,
            variables,
            objectives,
            perturbed_variables,
            perturbed_objectives,
            objective_weights,
            failed_realizations,
        )
        constraint_gradients = _calculate_estimated_constraint_gradients(
            self._config,
            self._function_estimators,
            variables,
            constraints,
            perturbed_variables,
            perturbed_constraints,
            constraint_weights,
            failed_realizations,
        )

        weighted_objective_gradient = np.array(
            (self._config.objectives.weights[:, np.newaxis] * objective_gradients).sum(
                axis=0
            )
        )

        return Gradients.create(
            weighted_objective=self._expand_gradients(
                weighted_objective_gradient, mask
            ),
            objectives=self._expand_gradients(objective_gradients, mask),
            constraints=(
                None
                if constraint_gradients is None
                else self._expand_gradients(constraint_gradients, mask)
            ),
        )

    @staticmethod
    def _expand_gradients(
        gradients: NDArray[np.float64],
        mask: NDArray[np.bool_] | None,
    ) -> NDArray[np.float64]:
        if mask is None:
            return gradients
        shape = gradients.shape[:-1] + (mask.size,)
        result = np.zeros(shape, dtype=np.float64)
        result[..., mask] = gradients
        return result

    def _calculate_filtered_realization_weights(
        self, evaluator_results: _FunctionEvaluatorResults
    ) -> tuple[NDArray[np.float64] | None, NDArray[np.float64] | None]:
        objective_weights: NDArray[np.float64] | None = None
        constraint_weights: NDArray[np.float64] | None = None

        objectives = evaluator_results.objectives
        assert objectives is not None
        constraints = evaluator_results.constraints

        objective_filters = self._config.objectives.realization_filters
        constraint_filters = (
            None
            if self._config.nonlinear_constraints is None
            else self._config.nonlinear_constraints.realization_filters
        )

        for idx, realization_filter in enumerate(self._realization_filters):
            apply_to_objectives = (
                None if objective_filters is None else objective_filters == idx
            )
            apply_to_constraints = (
                None if constraint_filters is None else constraint_filters == idx
            )
            if (apply_to_objectives is None or not np.any(apply_to_objectives)) and (
                apply_to_constraints is None or not np.any(apply_to_constraints)
            ):
                continue

            weights = realization_filter.get_realization_weights(
                objectives, constraints
            )
            if apply_to_objectives is not None:
                if objective_weights is None:
                    objective_weights = np.ones(
                        (
                            self._config.objectives.weights.size,
                            self._config.realizations.weights.size,
                        ),
                    )
                objective_weights[apply_to_objectives, :] = weights
            if constraint_filters is not None and apply_to_constraints is not None:
                assert self._config.nonlinear_constraints is not None
                if constraint_weights is None:
                    constraint_weights = np.ones(
                        (
                            self._config.nonlinear_constraints.lower_bounds.size,
                            self._config.realizations.weights.size,
                        ),
                    )
                constraint_weights[apply_to_constraints, :] = weights
        return objective_weights, constraint_weights

    def _init_realization_filters(
        self, plugin_manager: PluginManager
    ) -> list[RealizationFilter]:
        return [
            plugin_manager.get_plugin(
                "realization_filter", method=filter_config.method
            ).create(self._config, idx)
            for idx, filter_config in enumerate(self._config.realization_filters)
        ]

    def _init_function_estimators(
        self, plugin_manager: PluginManager
    ) -> list[FunctionEstimator]:
        return [
            plugin_manager.get_plugin(
                "function_estimator", method=estimator_config.method
            ).create(self._config, idx)
            for idx, estimator_config in enumerate(self._config.function_estimators)
        ]

    def _init_samplers(
        self, rng: Generator, plugin_manager: PluginManager
    ) -> list[Sampler]:
        samplers: list[Sampler] = []
        for idx, sampler_config in enumerate(self._config.samplers):
            variable_indices = _get_mask(
                idx, self._config.gradient.samplers, self._config.variables.mask
            )
            if variable_indices is None or variable_indices.size:
                plugin = plugin_manager.get_plugin(
                    "sampler", method=sampler_config.method
                )
                samplers.append(plugin.create(self._config, idx, variable_indices, rng))
        return samplers


def _get_mask(
    idx: int, gradient_indices: NDArray[np.intc] | None, mask: NDArray[np.bool_] | None
) -> NDArray[np.bool_] | None:
    if gradient_indices is None:
        return mask
    if mask is None:
        return np.asarray(gradient_indices == idx)
    return np.asarray(mask & (gradient_indices == idx))



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Any, TypeVar

import numpy as np

from ropt.enums import ResultAxis

TypeResultField = TypeVar("TypeResultField", bound="ResultField")


@dataclass(slots=True)
class ResultField:
    """Base class for `Results` fields."""

    @classmethod
    def get_axes(cls, name: str) -> tuple[ResultAxis, ...]:
        """Return the axes of a field in the given field class or object.

        When used with the class or an instance of that class for one of the
        fields of a `Results` object, retrieve the axes of the stored `numpy`
        array from the metadata and return them.

        Args:
            name: The name of the sub-field in the instance or class.

        Raises:
            ValueError: Raised if an unknown field name is passed.

        Returns:
            A tuple listing the names of the axes.
        """
        metadata = next(
            (item.metadata for item in fields(cls) if item.name == name), None
        )
        if metadata is None:
            msg = f"Unknown field name: {name}"
            raise ValueError(msg)
        return metadata.get("__axes__", ())

    def to_dict(
        self,
        name: str,
        axis: ResultAxis | None = None,
        names: dict[ResultAxis, tuple[str, ...]] | None = None,
    ) -> dict[str | int, Any]:
        """Convert a field of ResultsField to a dictionary.

        The keys of the output correspond to the indices of the axis of the
        field given by `axis`. The values are a slice of the field value at each
        index. If `names` contains the names of the objects stored along the axis,
        these are used as keys in the output dictionary rather than the numerical
        indices.

        If `axis` is `None` (the default), a default is chosen using the following logic:

        1. If the field has a `ResultAxis.OBJECTIVE` axis, it is used.
        2. If the field has a `ResultAxis.NONLINEAR_CONSTRAINT` axis, it is used.
        3. Otherwise, use the last axis.

        Args:
            name:  Name of the field to export.
            axis:  Axis to use as the keys.
            names: A dictionary mapping axis types to names.

        Returns:
            The field converted to a dictionary.
        """
        data = getattr(self, name)
        axes = self.get_axes(name)
        if axis is None:
            if ResultAxis.OBJECTIVE in axes:
                axis = ResultAxis.OBJECTIVE
            elif ResultAxis.NONLINEAR_CONSTRAINT in axes:
                axis = ResultAxis.NONLINEAR_CONSTRAINT
            else:
                axis = axes[-1]
        elif axis not in axes:
            msg = f"invalid key: {axis}"
            raise ValueError(msg)
        data = np.swapaxes(data, axes.index(axis), 0)
        if names is None:
            names = {}
        axis_names = names.get(axis, None)
        return (
            {idx: data[idx] for idx in range(data.shape[0])}
            if axis_names is None
            else {name: data[idx, ...] for idx, name in enumerate(axis_names)}
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from ropt.plugins import PluginManager
from ropt.plugins.optimizer.base import OptimizerCallback, OptimizerPlugin
from ropt.plugins.optimizer.scipy import SciPyOptimizerPlugin

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig


class TestPlugin(OptimizerPlugin):
    def create(self, _0: EnOptConfig, _1: OptimizerCallback) -> None:  # type: ignore[override]
        pass

    def is_supported(self, method: str) -> bool:
        return method.lower() in {"slsqp", "test"}


def test_default_plugins() -> None:
    plugin_manager = PluginManager()

    plugin = plugin_manager.get_plugin("optimizer", "slsqp")
    assert isinstance(plugin, SciPyOptimizerPlugin)


def test_default_plugins_full_spec() -> None:
    plugin_manager = PluginManager()

    plugin = plugin_manager.get_plugin("optimizer", "scipy/slsqp")
    assert isinstance(plugin, SciPyOptimizerPlugin)


def test_added_plugin() -> None:
    plugin_manager = PluginManager()
    plugin_manager.add_plugin("optimizer", "test", TestPlugin())

    plugin = plugin_manager.get_plugin("optimizer", "test")
    assert isinstance(plugin, TestPlugin)

    plugin = plugin_manager.get_plugin("optimizer", "slsqp")
    assert isinstance(plugin, SciPyOptimizerPlugin)

    plugin = plugin_manager.get_plugin("optimizer", "test/slsqp")
    assert isinstance(plugin, TestPlugin)


def test_added_plugin_prioritize() -> None:
    plugin_manager = PluginManager()
    plugin_manager.add_plugin("optimizer", "test", TestPlugin(), prioritize=True)

    plugin = plugin_manager.get_plugin("optimizer", "test")
    assert isinstance(plugin, TestPlugin)

    plugin = plugin_manager.get_plugin("optimizer", "slsqp")
    assert isinstance(plugin, TestPlugin)

    plugin = plugin_manager.get_plugin("optimizer", "scipy/slsqp")
    assert isinstance(plugin, SciPyOptimizerPlugin)



"""Configuration class for samplers."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict


class SamplerConfig(BaseModel):
    """The sampler configuration class.

    This class defines the configuration for samplers, which are configured by
    the `samplers` field in an [`EnOptConfig`][ropt.config.enopt.EnOptConfig]
    object. That field contains a tuple of configuration objects that define
    which samplers are available during the optimization. The `samplers` field
    in the gradient configuration
    ([`GradientConfig`][ropt.config.enopt.GradientConfig]) is used to specify
    the index of the sampler for each variable.

    Gradients are calculated from a set of perturbed variables, which may be
    deterministic or stochastic in nature. These perturbations are generally
    produced by sampler objects that produce perturbation values to add to the
    unperturbed variables.

    Perturbation values are produced by a sampler, which provides the methods
    that can be used. The `method` field determines which sampler method will be
    used. To further specify how such a method should function, the `options`
    field can be used to pass a dictionary of key-value pairs. The
    interpretation of these options depends on the chosen method.

    By default, a different set of perturbed variables is generated for each
    realization. By setting the `shared` flag to `True`, the sampler can be
    directed to use the same set of perturbed values for each realization.

    Attributes:
        method:  The sampler method.
        options: Options to be passed to the sampler.
        shared:  Whether perturbation values should be shared between realizations
                 (default: `False`).
    """

    method: str = "scipy/default"
    options: dict[str, Any] = {}
    shared: bool = False

    model_config = ConfigDict(
        extra="forbid",
        str_min_length=1,
        str_strip_whitespace=True,
        validate_default=True,
        frozen=True,
    )



from typing import Any, overload

from numpy.typing import NDArray


@overload
def _immutable_copy(data: NDArray[Any]) -> NDArray[Any]: ...


@overload
def _immutable_copy(data: NDArray[Any] | None) -> NDArray[Any] | None: ...


def _immutable_copy(data: NDArray[Any] | None) -> NDArray[Any] | None:
    if data is not None:
        data = data.copy()
        data.setflags(write=False)
    return data



import copy
from typing import Any

import numpy as np
import pytest

from ropt.config.enopt import EnOptConfig, GradientConfig, LinearConstraintsConfig
from ropt.enums import BoundaryType, PerturbationType
from ropt.transforms import OptModelTransforms, VariableScaler


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "variables": {
            "initial_values": np.array([1, 2]),
        },
        "objectives": {
            "weights": [1.0],
        },
        "optimizer": {
            "method": "dummy",
        },
    }


def test_check_linear_constraints() -> None:
    config = {
        "coefficients": np.array([[1, 2], [1, 2], [1, 2]]),
        "lower_bounds": np.array([-np.inf, 2, 3]),
        "upper_bounds": np.array([1, np.inf, 3]),
    }
    linear_constraints = LinearConstraintsConfig.model_validate(config)
    assert linear_constraints.coefficients is not None
    with pytest.raises(ValueError):  # noqa: PT011
        linear_constraints.coefficients[0, 0] = 0
    with pytest.raises(ValueError):  # noqa: PT011
        linear_constraints.upper_bounds[0] = 1


def test_check_linear_constraints_convert() -> None:
    config = {
        "coefficients": [[1, 2], [1, 2], [1, 2]],
        "lower_bounds": np.array([-np.inf, 2, 3]),
        "upper_bounds": np.array([1, np.inf, 3]),
    }
    LinearConstraintsConfig.model_validate(config)


def test_check_linear_constraints_vector_shapes() -> None:
    config = {
        "coefficients": [[1, 2, 3], [1, 2, 3]],
        "lower_bounds": np.array([-np.inf, 2]),
        "upper_bounds": np.array([1, np.inf]),
    }
    LinearConstraintsConfig.model_validate(config)

    config_copy = copy.deepcopy(config)
    config_copy["lower_bounds"] = [1, 2, 3]
    with pytest.raises(
        ValueError,
        match="lower_bounds cannot be broadcasted to a length of 2",
    ):
        LinearConstraintsConfig.model_validate(config_copy)


def test_check_perturbations() -> None:
    GradientConfig()
    gradients = GradientConfig(perturbation_magnitudes=np.array([0.1]))
    assert gradients.perturbation_magnitudes == np.array([0.1])


def test_check_config(enopt_config: Any) -> None:
    EnOptConfig.model_validate(enopt_config)


def test_check_config_linear_constraints(enopt_config: Any) -> None:
    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 2, 3], [2, 3, 4]],
        "lower_bounds": [1, 2],
        "upper_bounds": [np.inf, np.inf],
    }
    with pytest.raises(
        ValueError,
        match="the coefficients matrix should have 2 columns",
    ):
        EnOptConfig.model_validate(enopt_config)


def test_check_config_perturbations(enopt_config: Any) -> None:
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [1] * 2,
        "boundary_types": [BoundaryType.TRUNCATE_BOTH] * 2,
        "perturbation_types": [PerturbationType.ABSOLUTE] * 2,
    }
    EnOptConfig.model_validate(enopt_config)

    config_copy = copy.deepcopy(enopt_config)
    config_copy["gradient"]["perturbation_magnitudes"] = [1] * 3
    with pytest.raises(
        ValueError,
        match="the perturbation magnitudes cannot be broadcasted to a length of 2",
    ):
        EnOptConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(enopt_config)
    config_copy["gradient"]["boundary_types"] = [BoundaryType.TRUNCATE_BOTH] * 3
    with pytest.raises(
        ValueError, match="perturbation boundary_types must have 2 items"
    ):
        EnOptConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(enopt_config)
    config_copy["gradient"]["perturbation_types"] = [PerturbationType.ABSOLUTE] * 3
    with pytest.raises(ValueError, match="perturbation types must have 2 items"):
        EnOptConfig.model_validate(config_copy)


def test_check_config_min_success(enopt_config: Any) -> None:
    def gen_config(pert_min: int | None, real_min: int | None) -> dict[str, Any]:
        config: dict[str, Any] = copy.deepcopy(enopt_config)
        config["realizations"] = {"weights": 4 * [1.0]}
        config["gradient"] = {}
        if pert_min is not None:
            config["gradient"]["perturbation_min_success"] = pert_min
        if real_min is not None:
            config["realizations"]["realization_min_success"] = real_min
        return config

    pert_test_map = {None: 5, 1: 1, 4: 4, 7: 5}
    real_test_map = {None: 4, 1: 1, 3: 3, 7: 4}
    test_space = zip(pert_test_map.keys(), real_test_map.keys(), strict=False)
    for pert_in, real_in in test_space:
        config = EnOptConfig.model_validate(gen_config(pert_in, real_in))
        assert pert_test_map[pert_in] == config.gradient.perturbation_min_success
        assert real_test_map[real_in] == config.realizations.realization_min_success


def test_perturbation_types(enopt_config: Any) -> None:
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [0.1, 0.01],
        "perturbation_types": [PerturbationType.ABSOLUTE, PerturbationType.RELATIVE],
    }
    enopt_config["variables"]["lower_bounds"] = [0.0, 600]
    enopt_config["variables"]["upper_bounds"] = [1.0, np.inf]
    with pytest.raises(
        ValueError,
        match="The variable bounds must be finite to use relative perturbations",
    ):
        config = EnOptConfig.model_validate(enopt_config)

    enopt_config["variables"]["initial_values"] = [0.0, 0.0, 0.0]
    enopt_config["variables"]["lower_bounds"] = [0.0, 100.0, 0.0]
    enopt_config["variables"]["upper_bounds"] = [np.inf, 600.0, 1.0]
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [0.1, 0.01, 1.0],
        "perturbation_types": [
            PerturbationType.ABSOLUTE,
            PerturbationType.RELATIVE,
            PerturbationType.ABSOLUTE,
        ],
    }
    config = EnOptConfig.model_validate(enopt_config)
    assert np.allclose(config.gradient.perturbation_magnitudes, [0.1, 5.0, 1.0])


def test_perturbation_types_with_scaler(enopt_config: Any) -> None:
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [0.1, 0.01],
        "perturbation_types": [PerturbationType.ABSOLUTE, PerturbationType.RELATIVE],
    }
    enopt_config["variables"]["lower_bounds"] = [0.0, 600]
    enopt_config["variables"]["upper_bounds"] = [1.0, np.inf]
    with pytest.raises(
        ValueError,
        match="The variable bounds must be finite to use relative perturbations",
    ):
        config = EnOptConfig.model_validate(enopt_config)

    enopt_config["variables"]["initial_values"] = [0.0, 0.0, 0.0]
    enopt_config["variables"]["lower_bounds"] = [0.0, 100.0, 0.0]
    enopt_config["variables"]["upper_bounds"] = [np.inf, 600.0, 1.0]

    enopt_config["gradient"] = {
        "perturbation_magnitudes": [0.1, 0.01, 1.0],
        "perturbation_types": [
            PerturbationType.ABSOLUTE,
            PerturbationType.RELATIVE,
            PerturbationType.ABSOLUTE,
        ],
    }
    config = EnOptConfig.model_validate(
        enopt_config,
        context=OptModelTransforms(
            variables=VariableScaler(np.array([1.0, 1.0, 50.0]), None)
        ),
    )
    assert np.allclose(config.gradient.perturbation_magnitudes, [0.1, 5.0, 0.02])



"""Plugin functionality for adding function estimators.

This package contains the abstract base classes for function estimator plugins,
and the default function estimators that are part of `ropt`.
"""



"""The optimization event class."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig
    from ropt.enums import EventType


@dataclass(slots=True)
class Event:
    """The `Event` class stores optimization event data.

    While running an optimization plan, callbacks can be connected to react to
    events triggered during execution. These callbacks accept a single `Event`
    object containing information about the event.

    The actual data contained in the object depends on the nature of the event.
    Refer to the documentation of the [`EventType`][ropt.enums.EventType]
    enumeration for more details.

    Attributes:
        event_type: The type of the event.
        tags:       Optional tags added to the results.
        plan_id:    The ID of the plan that generated the event.
        data:       Optional data passed with the event.
    """

    event_type: EventType
    config: EnOptConfig
    tags: set[str] = field(default_factory=set)
    plan_id: tuple[int, ...] = field(default_factory=tuple)
    data: dict[str, Any] = field(default_factory=dict)



from typing import Any

import numpy as np
import pytest

from ropt.enums import OptimizerExitCode
from ropt.plan import BasicOptimizer
from ropt.results import FunctionResults, Results

pytestmark = [pytest.mark.slow]


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "optimizer": {
            "method": "external/slsqp",
            "tolerance": 1e-4,
            "max_iterations": 50,
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_external_run(enopt_config: Any, evaluator: Any) -> None:
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0, 0, 0.5], atol=0.02)


def test_external_max_functions_exceeded(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 0

    def track_results(_: tuple[Results, ...]) -> None:
        nonlocal last_evaluation

        last_evaluation += 1

    max_functions = 2
    enopt_config["optimizer"]["max_functions"] = max_functions
    optimizer = BasicOptimizer(enopt_config, evaluator()).set_results_callback(
        track_results
    )
    optimizer.run()
    assert last_evaluation == max_functions + 1
    assert optimizer.exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED


def test_external_failed_realizations(enopt_config: Any, evaluator: Any) -> None:
    def _observer(results: tuple[Results, ...]) -> None:
        assert isinstance(results[0], FunctionResults)
        assert results[0].functions is None

    functions = [lambda _0, _1: np.array(1.0), lambda _0, _1: np.array(np.nan)]
    optimizer = BasicOptimizer(enopt_config, evaluator(functions)).set_results_callback(
        _observer
    )
    optimizer.run()
    assert optimizer.exit_code == OptimizerExitCode.TOO_FEW_REALIZATIONS


def test_external_user_abort(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 0

    def _abort() -> bool:
        nonlocal last_evaluation

        if last_evaluation == 1:
            return True
        last_evaluation += 1
        return False

    optimizer = BasicOptimizer(enopt_config, evaluator()).set_abort_callback(_abort)
    optimizer.run()
    assert optimizer.results is not None
    assert last_evaluation == 1
    assert optimizer.exit_code == OptimizerExitCode.USER_ABORT


def test_external_error(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["optimizer"]["options"] = {"ftol": "foo"}
    with pytest.raises(
        RuntimeError,
        match="External optimizer error: could not convert string to float: 'foo'",
    ):
        BasicOptimizer(enopt_config, evaluator()).run()



from __future__ import annotations

from dataclasses import fields
from typing import TYPE_CHECKING, Any, Iterable, Sequence, cast

import pandas as pd

if TYPE_CHECKING:
    from ropt.enums import ResultAxis

    from ._result_field import ResultField


def _to_dataframe(  # noqa: PLR0913
    result_field: ResultField,
    plan_id: tuple[int, ...],
    batch_id: int | None,
    select: Iterable[str] | None,
    unstack: Iterable[ResultAxis] | None,
    names: dict[str, Sequence[str | int] | None] | None = None,
) -> pd.DataFrame:
    if select is None:
        select = (field.name for field in fields(result_field))
    if unstack is None:
        unstack = []
    joined_frame = pd.DataFrame()
    for field in select:
        series = _to_series(result_field, plan_id, batch_id, field, names)
        if series is not None:
            frame = series.to_frame()
            for axis in unstack:
                if axis.value in frame.index.names:
                    frame = cast(
                        pd.DataFrame,
                        frame.unstack(axis.value, sort=False),  # type:ignore[call-arg]  # noqa: PD010
                    )
            frame.columns = frame.columns.to_flat_index()  # type:ignore[no-untyped-call]
            if joined_frame.empty:
                joined_frame = frame
            else:
                joined_frame = joined_frame.join(frame, how="inner")
    return joined_frame


def _to_series(
    result_field: ResultField,
    plan_id: tuple[int, ...],
    batch_id: int | None,
    field: str,
    names: dict[str, Sequence[str | int] | None] | None = None,
) -> pd.Series[Any] | None:
    try:
        data = getattr(result_field, field)
    except AttributeError as exc:
        msg = f"Not a field name: {field}"
        raise ValueError(msg) from exc
    if data is None:
        return None
    axes = result_field.get_axes(field)
    if names is None:
        names = {}
    indices = [
        pd.RangeIndex(data.shape[idx]) if index is None else index
        for idx, index in enumerate(names.get(axis, None) for axis in axes)
    ]
    series: pd.Series[Any]
    index: tuple[Any, ...] = (plan_id, 0 if batch_id is None else batch_id)
    index_names = ["plan_id", "batch_id"]
    if indices:
        multi_index = pd.MultiIndex.from_product(
            indices, names=(axis.value for axis in axes)
        )
        series = pd.Series(data.flatten(), index=multi_index, name=field)
        series = pd.concat({index: series}, names=index_names)
    else:
        series = pd.Series(
            data,
            index=pd.MultiIndex.from_tuples([index], names=index_names),
            name=field,
        )
    return series



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeVar

from ._results import Results

if TYPE_CHECKING:
    from ropt.transforms._transforms import OptModelTransforms

    from ._gradient_evaluations import GradientEvaluations
    from ._gradients import Gradients
    from ._realizations import Realizations

TypeResults = TypeVar("TypeResults", bound="Results")


@dataclass(slots=True)
class GradientResults(Results):
    """The `GradientResults` class stores gradient related results.

    This contains  the following additional information:

    1. The results of the function evaluations for perturbed variables.
    2. The parameters of the realizations, such as weights for objectives and
       constraints, and realization failures.
    3. The gradients of the calculated objectives and constraints.

    Attributes:
        evaluations:  Results of the function evaluations.
        realizations: The calculated parameters of the realizations.
        gradients:    The calculated gradients.
    """

    evaluations: GradientEvaluations
    realizations: Realizations
    gradients: Gradients | None

    def transform_from_optimizer(
        self, transforms: OptModelTransforms
    ) -> GradientResults:
        """Apply transformations from optimizer space.

        Args:
            transforms: The transforms to apply.

        Returns:
            The transformed results.
        """
        return GradientResults(
            plan_id=self.plan_id,
            batch_id=self.batch_id,
            metadata=self.metadata,
            evaluations=self.evaluations.transform_from_optimizer(transforms),
            realizations=self.realizations,
            gradients=(
                None
                if self.gradients is None
                else self.gradients.transform_from_optimizer(transforms)
            ),
        )



"""This module defines a basic variable scaling transform."""

import numpy as np
from numpy.typing import NDArray

from .base import VariableTransform


class VariableScaler(VariableTransform):
    """Transform class for scaling variables."""

    def __init__(
        self, scales: NDArray[np.float64] | None, offsets: NDArray[np.float64] | None
    ) -> None:
        """Initialize the scaler.

        This scaler applies a linear scaling to the variables, defined
        by a scaling vector and an offset vector.

        if `scales` and `offsets` are both not `None`, they are broadcasted
        to the same length.

        Args:
            scales:  The scaling factors.
            offsets: The offset values.
        """
        if scales is not None and offsets is not None:
            scales, offsets = np.broadcast_arrays(scales, offsets)
        self._scales = scales
        self._offsets = offsets
        self._equation_scaling: NDArray[np.float64] | None = None

    def to_optimizer(self, values: NDArray[np.float64]) -> NDArray[np.float64]:
        """Implement the scaling to optimizer space.

        The values may consist of an array with multiple dimensions. It is
        assumed that the last axis contains the variable values. Should this
        method be used in a context where that is not the case, adjust the
        order of the axis accordingly before and after calling this method.

        Args:
            values: The values to be scaled.

        Returns:
            The scaled values.
        """
        if self._offsets is not None:
            values = values - self._offsets
        if self._scales is not None:
            values = values / self._scales
        return values

    def from_optimizer(self, values: NDArray[np.float64]) -> NDArray[np.float64]:
        """Implement the scaling from optimizer space.

        The values may consist of an array with multiple dimensions. It is
        assumed that the last axis contains the variable values. Should this
        method be used in a context where that is not the case, adjust the
        order of the axis accordingly before and after calling this method.

        Args:
            values: The values to be scaled.

        Returns:
            The scaled values.
        """
        if self._scales is not None:
            values = values * self._scales
        if self._offsets is not None:
            values = values + self._offsets
        return values

    def magnitudes_to_optimizer(
        self, values: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Implement the transformation of perturbation magnitudes.

        Args:
            values: The values to be transformed.

        Returns:
            The transformed values.
        """
        if self._scales is not None:
            return values / self._scales
        return values

    def linear_constraints_to_optimizer(
        self,
        coefficients: NDArray[np.float64],
        lower_bounds: NDArray[np.float64],
        upper_bounds: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """Implement the transformation of linear constraints.

        Args:
            coefficients: The coefficient matrix of the linear constraints.
            lower_bounds: The lower bounds on the right-hand-sides of the constraint equations.
            upper_bounds: The upper bounds on the right-hand-sides of the constraint equations.

        Returns:
            The transformed coefficient matrix and right-hand side bounds.
        """
        if self._offsets is not None:
            offsets = np.matmul(coefficients, self._offsets)
            lower_bounds = lower_bounds - offsets
            upper_bounds = upper_bounds - offsets
        if self._scales is not None:
            coefficients = coefficients * self._scales
        self._equation_scaling = np.max(np.abs(coefficients), axis=-1)
        assert self._equation_scaling is not None
        return (
            coefficients / self._equation_scaling[:, np.newaxis],
            lower_bounds / self._equation_scaling,
            upper_bounds / self._equation_scaling,
        )

    def bound_constraint_diffs_from_optimizer(
        self, lower_diffs: NDArray[np.float64], upper_diffs: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Implement the transformation of bound constraint diffs from optimizer space.

        Args:
            lower_diffs: The lower differences to be transformed.
            upper_diffs: The upper differences to be transformed.

        Returns:
            The transformed values.
        """
        if self._scales is not None:
            lower_diffs = lower_diffs * self._scales
            upper_diffs = upper_diffs * self._scales
        return lower_diffs, upper_diffs

    def linear_constraints_diffs_from_optimizer(
        self, lower_diffs: NDArray[np.float64], upper_diffs: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Implement the transformation of bound constraint diffs from optimizer space.

        Args:
            lower_diffs: The lower differences to be transformed.
            upper_diffs: The upper differences to be transformed.

        Returns:
            The transformed values.
        """
        if self._equation_scaling is not None:
            lower_diffs = lower_diffs * self._equation_scaling
            upper_diffs = upper_diffs * self._equation_scaling
        return lower_diffs, upper_diffs



from functools import partial
from typing import Any, Sequence

import pytest

from ropt.config.enopt import EnOptConfig
from ropt.enums import ResultAxis
from ropt.plan import BasicOptimizer
from ropt.report import ResultsDataFrame
from ropt.results import Results

# Requires pandas:
pd = pytest.importorskip("pandas")


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "speculative": True,
            "max_functions": 3,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
            "upper_bounds": 1.0,
            "lower_bounds": -1.0,
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def _handle_results(
    results: tuple[Results, ...],
    reporter: ResultsDataFrame,
    variable_names: Sequence[str | int] | None = None,
) -> None:
    names: dict[str, Sequence[str | int] | None] | None = (
        None if variable_names is None else {ResultAxis.VARIABLE: variable_names}
    )
    for item in results:
        reporter.add_results(item, names=names)


def test_dataframe_results_no_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(set())
    BasicOptimizer(config, evaluator()).set_results_callback(
        partial(_handle_results, reporter=reporter),
    ).run()
    assert reporter.frame.empty


def test_dataframe_results_function_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "evaluations.variables",
        },
    )
    BasicOptimizer(config, evaluator()).set_results_callback(
        partial(_handle_results, reporter=reporter),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", idx) for idx in range(3)
    ]


def test_dataframe_results_function_results_formatted_names(
    enopt_config: Any, evaluator: Any
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "evaluations.variables",
        },
    )
    BasicOptimizer(config, evaluator()).set_results_callback(
        partial(
            _handle_results,
            reporter=reporter,
            variable_names=[f"a:{idx}" for idx in range(1, 4)],
        ),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", f"a:{idx}") for idx in range(1, 4)
    ]


def test_dataframe_results_gradient_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "gradients.weighted_objective",
        },
        table_type="gradients",
    )
    BasicOptimizer(config, evaluator()).set_results_callback(
        partial(
            _handle_results,
            reporter=reporter,
            variable_names=[f"a:{idx}" for idx in range(1, 4)],
        ),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("gradients.weighted_objective", f"a:{idx}") for idx in range(1, 4)
    ]


def test_dataframe_results_metadata(enopt_config: Any, evaluator: Any) -> None:
    reporter = ResultsDataFrame(
        {
            "evaluations.variables",
            "metadata.foo.bar",
            "metadata.not.existing",
        },
    )

    def even_handler(results: tuple[Results, ...]) -> None:
        for item in results:
            item.metadata["foo"] = {"bar": 1}
        for item in results:
            reporter.add_results(item)

    BasicOptimizer(enopt_config, evaluator()).set_results_callback(even_handler).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", idx) for idx in range(3)
    ] + ["metadata.foo.bar"]



"""This module defines the OptModelTransforms class."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base import (
        NonLinearConstraintTransform,
        ObjectiveTransform,
        VariableTransform,
    )


@dataclass
class OptModelTransforms:
    """A container for optimization model transformers."""

    variables: VariableTransform | None = None
    objectives: ObjectiveTransform | None = None
    nonlinear_constraints: NonLinearConstraintTransform | None = None



"""This module defines the optimization plan object."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from ropt.exceptions import PlanAborted
from ropt.plugins.plan.base import PlanStep, ResultHandler

if TYPE_CHECKING:
    from ropt.plan import Event

    from ._context import OptimizerContext


class Plan:
    """The plan class for executing optimization workflows."""

    def __init__(
        self,
        optimizer_context: OptimizerContext,
        parent: Plan | None = None,
        plan_id: tuple[int, ...] | None = None,
    ) -> None:
        """Initialize a plan object.

        This method initializes a plan using
        an [`OptimizerContext`][ropt.plan.OptimizerContext] object. An optional
        [`plugin_manager`][ropt.plugins.PluginManager] argument allows for the
        specification of custom plugins for result handlers and step objects
        within the plan. If omitted, only plugins installed through Python's
        standard entry points are used.

        Plan hierarchies are tracked using the `plan_id` attribute, a tuple of
        integers. By default a plan has `plan_id == (0,)`. A plan may set a
        parent plan using the `set_parent` method, which may add an new index at
        the end of the `plan_id` tuple. This structure enables efficient tracing
        across both sequential and nested plan workflows.

        Args:
            optimizer_context: The context in which the plan will execute,
                               providing shared resources across steps.
            parent:            Optional reference to a parent plan.
            plan_id:           The ID of the plan, reflecting its hierarchy.
        """
        self._optimizer_context = optimizer_context
        self._vars: dict[str, Any] = {}
        self._plan_id: tuple[int, ...] = (0,) if plan_id is None else plan_id

        self._aborted = False
        self._parent = parent
        self._handlers: list[ResultHandler] = []
        self["plan_id"] = list(self.plan_id)
        self._function: Callable[..., Any] | None = None

    def add_handler(self, name: str, **kwargs: Any) -> ResultHandler:  # noqa: ANN401
        """Add a handler to the plan.

        Args:
            name: The name of the handler to add.
            kwargs: Additional arguments to pass to the handler.

        Returns:
            The new handler.
        """
        handler = self._optimizer_context.plugin_manager.get_plugin(
            "plan_handler", method=name
        ).create(name, self, **kwargs)
        self._handlers.append(handler)
        assert isinstance(handler, ResultHandler)
        return handler

    def handler_exists(self, name: str) -> bool:
        """Check if a handler exists.

        Args:
            name: The name of the handler to check.

        Returns:
            Whether the handler exists.
        """
        return self._optimizer_context.plugin_manager.is_supported(
            "plan_handler", method=name
        )

    def clear_handlers(self) -> None:
        """Clear all handlers from the plan."""
        self._handlers.clear()

    def add_step(self, name: str, **kwargs: Any) -> PlanStep:  # noqa: ANN401
        """Add a step to the plan.

        Args:
            name: The name of the step to run.
            kwargs: Additional arguments to pass to the step.

        Returns:
            The step that was executed.
        """
        step = self._optimizer_context.plugin_manager.get_plugin(
            "plan_step", method=name
        ).create(name, self, **kwargs)
        assert isinstance(step, PlanStep)
        return step

    def step_exists(self, name: str) -> bool:
        """Check if a step exists.

        Args:
            name: The name of the step to check.

        Returns:
            Whether the step exists.
        """
        return self._optimizer_context.plugin_manager.is_supported(
            "plan_step", method=name
        )

    def add_function(self, func: Callable[..., Any]) -> None:
        """Add a function to the plan.

        The function can be called using the `run_function` method, and is
        generally used to execute one or more steps in the plan.

        Args:
            func: The function to add to the plan.
        """
        self._function = func

    def run_step(self, step: PlanStep, **kwargs: Any) -> None:  # noqa: ANN401
        """Run a step in the plan.

        Args:
            step:   The step to run
            kwargs: Additional arguments to pass to the step.

        Raises:
            PlanAborted: _description_
        """
        if self._aborted:
            msg = "Plan was aborted by the previous step."
            raise PlanAborted(msg)
        step.run(**kwargs)

    def run_function(self, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        """Run a function in the plan.

        Args:
            args:   Arbitrary positional arguments to pass to the function.
            kwargs: Arbitrary keyword arguments to pass to the function.

        Returns:
            The result of the function.
        """
        if self._function is None:
            msg = "No function has been added to the plan."
            raise AttributeError(msg)
        return self._function(self, *args, **kwargs)

    def abort(self) -> None:
        """Abort the plan.

        This will not abort a running plan, but will prevent any further steps
        from being executed. This can be used to stop the plan from continuing
        if a step fails or if the user decides to stop the optimization.

        The `aborted` method can be used to check if the plan was aborted.
        """
        self._aborted = True

    @property
    def aborted(self) -> bool:
        """Check if the plan was aborted.

        Returns:
            bool: `True` if the plan was aborted; otherwise, `False`.
        """
        return self._aborted

    @property
    def plan_id(self) -> tuple[int, ...]:
        """Return the ID of the plan.

        Each plan has a unique ID, stored as a tuple of integers, which reflects
        creation order and any nesting structure within the plan
        hierarchy.

        Returns:
            tuple: The unique tuple-based ID for this plan.
        """
        return self._plan_id

    @property
    def optimizer_context(self) -> OptimizerContext:
        """Return the optimizer context associated with the plan.

        This method retrieves the optimizer context object that provides shared
        state and functionality for executing the optimization plan.

        Returns:
            The optimizer context object used by the plan.
        """
        return self._optimizer_context

    def set_parent(self, parent: Plan, add_id: int) -> None:
        self._plan_id = (*parent.plan_id, add_id)
        self._parent = parent

    def emit_event(self, event: Event) -> None:
        """Emit an event of the specified type with the provided data.

        When this method is called, the following steps are executed:

        1. All event handlers associated with the plan are invoked, which may
           modify the event.
        2. If the plan has no parent, all observer functions registered for the
           specified event type via the `add_observer` method are called.
        3. If the plan has a parent, the `emit_event` method of the parent plan
           is also called.

        Args:
            event: The event object to emit.
        """
        if not event.plan_id:
            event.plan_id = self.plan_id
        for handler in self._handlers:
            event = handler.handle_event(event)
        if self._parent is None:
            self._optimizer_context.call_observers(event)
        else:
            self._parent.emit_event(event)

    def __getitem__(self, name: str) -> Any:  # noqa: ANN401
        """Get the value of a plan variable.

        This method implements the `[]` operator on the plan object to retrieve
        the value associated with a specific plan variable.

        Args:
            name: The name of the variable whose value is to be retrieved.

        Returns:
            The value of the specified variable, which can be of any type.
        """
        if name in self._vars:
            return self._vars[name]
        msg = f"Unknown plan variable: `{name}`"
        raise AttributeError(msg)

    def __setitem__(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Set a plan variable to the given value.

        This method implements the `[]` operator on the plan object to set the
        value of a specific plan variable.

        Args:
            name:  The name of the variable to set.
            value: The value to assign to the variable.
        """
        if not name.isidentifier():
            msg = f"Not a valid variable name: `{name}`"
            raise AttributeError(msg)
        self._vars[name] = value

    def __contains__(self, name: str) -> bool:
        """Check if a variable exists.

        This method implements the `in` operator on the plan object to determine
        if a plan variable exists.

        Args:
            name: The name of the variable.

        Returns:
            `True` if the variable exists; otherwise, `False`.
        """
        return name in self._vars



"""The `ropt.config.enopt` module contains optimization configuration classes."""

from ._enopt_config import EnOptConfig
from ._function_estimator_config import FunctionEstimatorConfig
from ._gradient_config import GradientConfig
from ._linear_constraints_config import LinearConstraintsConfig
from ._nonlinear_constraints_config import NonlinearConstraintsConfig
from ._objective_functions_config import ObjectiveFunctionsConfig
from ._optimizer_config import OptimizerConfig
from ._realization_filter_config import RealizationFilterConfig
from ._realizations_config import RealizationsConfig
from ._sampler_config import SamplerConfig
from ._variables_config import VariablesConfig

__all__ = [
    "EnOptConfig",
    "FunctionEstimatorConfig",
    "GradientConfig",
    "LinearConstraintsConfig",
    "NonlinearConstraintsConfig",
    "ObjectiveFunctionsConfig",
    "OptimizerConfig",
    "RealizationFilterConfig",
    "RealizationsConfig",
    "SamplerConfig",
    "VariablesConfig",
]



"""Configuration class for variables."""

from __future__ import annotations

from typing import Self

import numpy as np
from pydantic import ConfigDict, ValidationInfo, model_validator

from ropt.config.utils import (
    ImmutableBaseModel,
    broadcast_1d_array,
    check_enum_values,
    immutable_array,
)
from ropt.config.validated_types import (  # noqa: TC001
    Array1D,
    Array1DBool,
    ArrayEnum,
)
from ropt.enums import VariableType


class VariablesConfig(ImmutableBaseModel):
    r"""The configuration class for variables.

    This configuration class, configured by the `variables` field in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object defines essential
    aspects of the variables: the initial values and the bounds. These are given
    by the `initial_values`, `lower_bounds`, and `upper_bounds` fields, which
    are [`numpy`](https://numpy.org) arrays. Initial values must be provided,
    and its length determines the number of variables. The lower and upper
    bounds, are broadcasted to the number of variables, and are set to
    $-\infty$ and $+\infty$ by default. They may contain `numpy.nan` values,
    indicating that corresponding variables have no lower or upper bounds,
    respectively. These values are converted to `numpy.inf` values with an
    appropriate sign.

    The optional `types` field can be used to assign types to each variable,
    according to the [`VariableType`][ropt.enums.VariableType] enumeration. The
    values can be used to configure the optimizer accordingly. If not provided,
    all variables are assumed to be continuous and of real data type
    (corresponding to [`VariableType.REAL`][ropt.enums.VariableType.REAL])

    The boolean entries of the optional `mask` field indicates which variables
    are considered to be free to change. During optimization, only these
    variables should change while others remain fixed.

    Attributes:
        types:          The type of the variables (optional).
        initial_values: The initial values of the variables.
        lower_bounds:   Lower bound of the variables (default: $-\infty$).
        upper_bounds:   Upper bound of the variables (default: $+\infty$).
        mask:           Optional mask of variables to optimize.
    """

    types: ArrayEnum | None = None
    initial_values: Array1D = np.array(0.0)
    lower_bounds: Array1D = np.array(-np.inf)
    upper_bounds: Array1D = np.array(np.inf)
    mask: Array1DBool | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _broadcast_and_transform(self, info: ValidationInfo) -> Self:
        self._mutable()

        lower_bounds = broadcast_1d_array(
            self.lower_bounds, "lower_bounds", self.initial_values.size
        )
        upper_bounds = broadcast_1d_array(
            self.upper_bounds, "upper_bounds", self.initial_values.size
        )

        if info.context is not None and info.context.variables is not None:
            self.initial_values = immutable_array(
                info.context.variables.to_optimizer(self.initial_values)
            )
            lower_bounds = info.context.variables.to_optimizer(lower_bounds)
            upper_bounds = info.context.variables.to_optimizer(upper_bounds)

        if np.any(lower_bounds > upper_bounds):
            msg = "The lower bounds are larger than the upper bounds."
            raise ValueError(msg)

        self.lower_bounds = immutable_array(lower_bounds)
        self.upper_bounds = immutable_array(upper_bounds)

        if self.types is not None:
            check_enum_values(self.types, VariableType)
            self.types = broadcast_1d_array(
                self.types, "types", self.initial_values.size
            )
        if self.mask is not None:
            self.mask = broadcast_1d_array(self.mask, "mask", self.initial_values.size)

        self._immutable()

        return self



"""Annotated types for use with Pydantic models.

These types can be used to convert input values to a desired type and guarantee
certain properties. They include types that convert inputs to immutable NumPy
arrays of specified dimension and type:

- [`Array1D`][ropt.config.validated_types.Array1D]: For converting sequences to
  immutable one-dimensional floating-point arrays.
- [`Array2D`][ropt.config.validated_types.Array2D]: For converting sequences to
  immutable two-dimensional floating-point arrays.
- [`ArrayEnum`][ropt.config.validated_types.ArrayEnum]: For converting
  sequences to values of numerical enumerations of any dimension.
- [`Array1DInt`][ropt.config.validated_types.Array1DInt]: For converting
  sequences to immutable one-dimensional integer arrays.
- [`Array1DBool`][ropt.config.validated_types.Array1DBool]: For converting
  sequences to immutable one-dimensional boolean arrays.

Additionally, the following convenience types create sets or tuples, ensuring
that single values are embedded in a set or tuple, respectively:

- [`ItemOrSet[T]`][ropt.config.validated_types.ItemOrSet]: Create a set of type `T`.
- [`ItemOrTuple[T]`][ropt.config.validated_types.ItemOrTuple]: Create a tuple of type `T`.
"""

from typing import Annotated, TypeVar

import numpy as np
from numpy.typing import NDArray
from pydantic import BeforeValidator

from .utils import (
    _convert_1d_array,
    _convert_1d_array_bool,
    _convert_1d_array_intc,
    _convert_2d_array,
    _convert_enum_array,
    _convert_set,
    _convert_tuple,
)

Array1D = Annotated[NDArray[np.float64], BeforeValidator(_convert_1d_array)]
"""Convert to an immutable 1D numpy array of floating point values."""

Array2D = Annotated[NDArray[np.float64], BeforeValidator(_convert_2d_array)]
"""Convert to an immutable 2D numpy array of floating point values."""

ArrayEnum = Annotated[NDArray[np.ubyte], BeforeValidator(_convert_enum_array)]
"""Convert to an immutable numpy array of numerical enumeration values."""

Array1DInt = Annotated[NDArray[np.intc], BeforeValidator(_convert_1d_array_intc)]
"""Convert to an immutable 1D numpy array of integer values."""

Array1DBool = Annotated[NDArray[np.bool_], BeforeValidator(_convert_1d_array_bool)]
"""Convert to an immutable 1D numpy array of boolean values."""


T = TypeVar("T")
ItemOrSet = Annotated[set[T], BeforeValidator(_convert_set)]
"""Convert to single value to a set containing that value, passes sets unchanged."""

ItemOrTuple = Annotated[tuple[T, ...], BeforeValidator(_convert_tuple)]
"""Convert to single value to a tuple containing that value, passes sets unchanged."""



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ropt.enums import ResultAxis

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.transforms import OptModelTransforms


@dataclass(slots=True)
class Functions(ResultField):
    """Store the calculated objective and constraint functions.

    The objective and constraint functions used by the optimizer are calculated
    from their values evaluated for all realizations, for instance by averaging.
    There may be multiple objectives and constraints. Multiple objectives are
    handled by the optimizer using a weighted sum, stored in the
    `weighted_objective` field. Multiple constraints are directly handled by the
    optimizer.

    Attributes:
        weighted_objective: The weighted sum of the objectives.
        objectives:         The value of each objective.
        constraints:        The value of each constraint.
    """

    weighted_objective: NDArray[np.float64] = field(
        metadata={"__axes__": ()},
    )
    objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (ResultAxis.OBJECTIVE,),
        },
    )
    constraints: NDArray[np.float64] | None = field(
        default=None,
        metadata={
            "__axes__": (ResultAxis.NONLINEAR_CONSTRAINT,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.weighted_objective = _immutable_copy(self.weighted_objective)
        self.objectives = _immutable_copy(self.objectives)
        self.constraints = _immutable_copy(self.constraints)

    @classmethod
    def create(
        cls,
        weighted_objective: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: NDArray[np.float64] | None = None,
    ) -> Functions:
        """Create a Functions object with the given information.

        Args:
            config:                 Configuration object.
            weighted_objective:     The weighted objective.
            objectives:             The objective functions for each realization.
            constraints:            The constraint functions for each realization.

        Returns:
            A new Functions object.
        """
        return Functions(
            weighted_objective=weighted_objective,
            objectives=objectives,
            constraints=constraints,
        )

    def transform_from_optimizer(self, transforms: OptModelTransforms) -> Functions:
        """Apply transformations from optimizer space.

        Args:
            transforms: The transforms to apply.

        Returns:
            The transformed results.
        """
        if transforms.objectives is None and transforms.nonlinear_constraints is None:
            return self

        return Functions(
            weighted_objective=(
                self.weighted_objective
                if transforms.objectives is None
                else transforms.objectives.weighted_objective_from_optimizer(
                    self.weighted_objective
                )
            ),
            objectives=(
                self.objectives
                if transforms.objectives is None
                else transforms.objectives.from_optimizer(self.objectives)
            ),
            constraints=(
                self.constraints
                if self.constraints is None or transforms.nonlinear_constraints is None
                else transforms.nonlinear_constraints.from_optimizer(self.constraints)
            ),
        )



"""Configuration class for the optimizer."""

from __future__ import annotations

from pathlib import Path  # noqa: TC003
from typing import Any, Self

from pydantic import (
    ConfigDict,
    NonNegativeFloat,
    PositiveInt,
    model_validator,
)

from ropt.config.utils import ImmutableBaseModel


class OptimizerConfig(ImmutableBaseModel):
    """The configuration class for optimizers used in the optimization.

    This class defines the configuration for optimizers, configured by the
    `optimizer` field in an [`EnOptConfig`][ropt.config.enopt.EnOptConfig]
    object.

    Although there may be significant differences in the parameters that can be
    used for different optimization methods, there are a few standard settings
    defined in this configuration object, which are forwarded to the optimizer:

    - The maximum number of iterations allowed before the optimization should be
      aborted by this optimizer. The optimizer may choose to ignore this option.
    - The maximum number of function evaluations allowed before the optimization
      is aborted.
    - The convergence tolerance used as a stopping criterion. The exact
      definition of the criterion depends on the optimizer. The optimizer may
      choose to ignore this option.
    - Whether gradients should be evaluated early, even if the optimizer does
      not strictly need it yet. When evaluating on a distributed HPC cluster,
      this may lead to better load-balancing for some methods. This option is
      only applied if the optimization algorithm knows how to make use of it.
    - Whether calculations for functions and gradients should be done
      separately, even if the optimizer requests them to be evaluated together.
      This option is useful when a filter is specified that deactivates some
      realizations (see
      [`RealizationFilterConfig`][ropt.config.enopt.RealizationFilterConfig]).
      In this case, after evaluation of the functions, it may be possible to
      reduce the number of evaluations for a following gradient calculation.
    - Whether the optimizer may use parallelized function evaluations. This
      option currently only applies to gradient-free methods and may be ignored
      by the optimizer.
    - An optional location of an output directory, where the optimizer may store
      files.
    - Generic optimizer options that may be passed as an arbitrary dictionary,
      or as a list of strings. It depends on the method what form is required
      and how it is interpreted.

    Attributes:
        method:            Name of the optimization method used.
        max_iterations:    Optional maximum number of iterations.
        max_functions:     Optional maximum number of function evaluations.
        tolerance:         Optional tolerance for convergence.
        speculative:       Force gradient evaluations; disabled if
                           split_evaluations is True (default `False`).
        split_evaluations: Evaluate function and gradient separately
                           (default: `False`).
        parallel:          Allow for parallelized evaluation (default: `False`).
        output_dir:        Optional output directory for use by the optimizer.
        options:           Optional generic options for use by the optimizer.
    """

    method: str = "scipy/default"
    max_iterations: PositiveInt | None = None
    max_functions: PositiveInt | None = None
    tolerance: NonNegativeFloat | None = None
    speculative: bool = False
    split_evaluations: bool = False
    parallel: bool = False
    output_dir: Path | None = None
    options: dict[str, Any] | list[str] | None = None

    model_config = ConfigDict(
        extra="forbid",
        str_min_length=1,
        str_strip_whitespace=True,
        validate_default=True,
    )

    @model_validator(mode="after")
    def _method(self) -> Self:
        self._mutable()
        plugin, sep, method = self.method.rpartition("/")
        if (sep == "/") and (plugin == "" or method) == "":
            msg = f"malformed method specification: `{self.method}`"
            raise ValueError(msg)
        self._mutable()
        return self



"""Data classes for storing intermediate optimization results.

During the optimization process, functions and gradients are calculated and need
to be reported. To streamline this process, new results are passed to callbacks
as a tuple of [`Results`][ropt.results.Results] objects. These may be instances
of the derived [`FunctionResults`][ropt.results.FunctionResults] or
[`GradientResults`][ropt.results.GradientResults] classes, which contain results
for function and gradient evaluations, respectively.

Much of the data stored in the result objects is multi-dimensional. For example,
the `objectives` field, which is part of the nested
[`evaluations`][ropt.results.FunctionEvaluations] object in the
[`FunctionResults`][ropt.results.FunctionResults], is a two-dimensional `numpy`
array. In this array, columns correspond to the objectives, and rows correspond
to the realization numbers.

To facilitate exporting and reporting results, the identity of the axes in such
multi-dimensional arrays is stored in metadata associated with the corresponding
field. These fields derive from the [`ResultField`][ropt.results.ResultField]
class, which includes a [`get_axes`][ropt.results.ResultField.get_axes] class
method to retrieve the axes. For example, for the `objectives` field, this
method retrieves the axes:

```py
>>> from ropt.results import FunctionEvaluations
>>> FunctionEvaluations.get_axes("objectives")
(<ResultAxis.REALIZATION: 'realization'>, <ResultAxis.OBJECTIVE: 'objective'>)
```

Using this metadata, the exporting or reporting code can refer to the optimizer
configuration to associate realization and objective names with any entry in the
result matrix. For instance, the [`pandas`](https://pandas.pydata.org/)
exporting code will utilize this information to construct a multi-index for the
generated DataFrame and may also unstack such multi-dimensional data into
multiple columns.
"""

from ._constraint_info import ConstraintInfo
from ._function_evaluations import FunctionEvaluations
from ._function_results import FunctionResults
from ._functions import Functions
from ._gradient_evaluations import GradientEvaluations
from ._gradient_results import GradientResults
from ._gradients import Gradients
from ._realizations import Realizations
from ._result_field import ResultField
from ._results import Results

__all__ = [
    "ConstraintInfo",
    "FunctionEvaluations",
    "FunctionResults",
    "Functions",
    "GradientEvaluations",
    "GradientResults",
    "Gradients",
    "Realizations",
    "ResultField",
    "Results",
]



"""Generate a results report in a `pandas` data frame."""

from __future__ import annotations

from dataclasses import fields
from functools import partial
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Final, Literal, Sequence

from ropt.enums import ResultAxis
from ropt.results import FunctionResults, GradientResults

if TYPE_CHECKING:
    from ropt.results import Results

_HAVE_PANDAS: Final = find_spec("pandas") is not None

if TYPE_CHECKING:
    from ropt.results import Results

if _HAVE_PANDAS:
    import pandas as pd


class ResultsDataFrame:
    """Generate a results report in a `pandas` DataFrame.

    The class is intended to be used for gathering results from
    [`Results`][ropt.results.Results] objects and storing them as a
    [`pandas`](https://pandas.pydata.org/) DataFrame.

    New results can be added to the stored DataFrame as they become available
    using the `add_results` method. The content of the DataFrame is taken from
    the fields of the [`Results`][ropt.results.Results] object passed during
    each call to `add_results`. The updated table can be retrieved at any time
    via the [`frame`][ropt.report.ResultsDataFrame.frame] property.
    """

    def __init__(
        self,
        fields: set[str],
        *,
        table_type: Literal["functions", "gradients"] = "functions",
    ) -> None:
        """Initialize a ResultsDataFrame object.

        The set of fields used is determined by the `fields` argument passed
        when initializing the object. These can be any of the fields defined in
        a [`FunctionResults`][ropt.results.FunctionResults] or a
        [`GradientResults`][ropt.results.GradientResults]. Most of the fields in
        these objects are nested, and the fields to export must be specified
        specifically in the form `field.subfield`. For instance, to specify the
        `variables` field of the `evaluations` field from a function result, the
        specification would be `evaluations.variables`.

        Note that many fields may, in fact, generate multiple columns in the
        resulting data frame. For instance, when specifying
        `evaluations.variables`, a column will be generated for each variable.
        If available, variable names, will be used as column labels. Because the
        exported fields may be multi-dimensional with names defined along each
        axis, for instance, realizations and objectives, which both can be
        named, the final name may consist of a tuple of names.

        The `table_type` argument is used to determine which type of results
        should be reported: either function evaluation results (`functions`) or
        gradient results (`gradients`).

        Instead of passing a set of field specifiers via the `fields` argument,
        a dictionary mapping the field specifiers to column names can be passed.
        In this case the class will function exactly the same, using the set of
        keys of the dictionary as field specifiers. However, now the `get_table`
        method can be used to retrieve a data frame where all columns are sorted
        and renamed according to the dictionary's content.

        Args:
            fields:     The fields of the results to store.
            table_type: The type of the table.
        """
        if not _HAVE_PANDAS:
            msg = "ResultsDataFrame requires the `pandas` modules"
            raise NotImplementedError(msg)

        self._fields = fields
        self._table_type = table_type
        self._frame = pd.DataFrame()

    def add_results(
        self,
        results: Results,
        names: dict[str, Sequence[str | int] | None] | None = None,
    ) -> bool:
        """Add a results object to the table.

        This method can be called directly from any observers connected to
        events that produce results.

        The `names` argument is an optional dictionary that maps axis types to
        names, that are used to label the multi-index columns in the resulting
        data frame. If not provided, numerical indices are used.

        Args:
            results: The results to add.
            names:   A dictionary mapping axis types to names.

        Returns:
            True if a result was added, else False
        """
        if not isinstance(results, (FunctionResults, GradientResults)):
            msg = "ResultsDataFrame.add_results() requires FunctionResults or GradientResults"
            raise TypeError(msg)

        frame: pd.DataFrame | None = None
        if (
            self._table_type == "functions"
            and isinstance(results, FunctionResults)
            and results.functions is not None
        ):
            frame = _get_function_results(results, self._fields, names)
        elif (
            self._table_type == "gradients"
            and isinstance(results, GradientResults)
            and results.gradients is not None
        ):
            frame = _get_gradient_results(results, self._fields, names)
        if frame is not None:
            frame = _add_metadata(frame, results, self._fields)
            self._frame = pd.concat([self._frame, frame])
            return True

        return False

    @property
    def frame(self) -> pd.DataFrame:
        """Return the function results generated so far.

        Returns:
            A pandas data frame with the results.
        """
        return self._frame


def _get_function_results(
    results: Results,
    sub_fields: set[str],
    names: dict[str, Sequence[str | int] | None] | None,
) -> pd.DataFrame:
    if (
        not sub_fields
        or not isinstance(results, FunctionResults)
        or results.functions is None
    ):
        return pd.DataFrame()

    functions = results.to_dataframe(
        "functions",
        select=_get_select(results, "functions", sub_fields),
        unstack=[ResultAxis.OBJECTIVE, ResultAxis.NONLINEAR_CONSTRAINT],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="functions"))

    evaluations = results.to_dataframe(
        "evaluations",
        select=_get_select(results, "evaluations", sub_fields),
        unstack=[
            ResultAxis.VARIABLE,
            ResultAxis.OBJECTIVE,
            ResultAxis.NONLINEAR_CONSTRAINT,
        ],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="evaluations"))

    if results.constraint_info is not None:
        constraint_info = results.to_dataframe(
            "constraint_info",
            select=_get_select(results, "constraint_info", sub_fields),
            unstack=[
                ResultAxis.VARIABLE,
                ResultAxis.LINEAR_CONSTRAINT,
                ResultAxis.NONLINEAR_CONSTRAINT,
            ],
            names=names,
        ).rename(columns=partial(_add_prefix, prefix="constraint_info"))

        return _join_frames(functions, evaluations, constraint_info)

    return _join_frames(functions, evaluations)


def _get_gradient_results(
    results: Results,
    sub_fields: set[str],
    names: dict[str, Sequence[str | int] | None] | None,
) -> pd.DataFrame:
    if (
        not sub_fields
        or not isinstance(results, GradientResults)
        or results.gradients is None
    ):
        return pd.DataFrame()

    gradients = results.to_dataframe(
        "gradients",
        select=_get_select(results, "gradients", sub_fields),
        unstack=[
            ResultAxis.OBJECTIVE,
            ResultAxis.NONLINEAR_CONSTRAINT,
            ResultAxis.VARIABLE,
        ],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="gradients"))

    evaluations = results.to_dataframe(
        "evaluations",
        select=_get_select(results, "evaluations", sub_fields),
        unstack=[
            ResultAxis.VARIABLE,
            ResultAxis.OBJECTIVE,
            ResultAxis.NONLINEAR_CONSTRAINT,
        ],
        names=names,
    ).rename(columns=partial(_add_prefix, prefix="evaluations"))

    return _join_frames(gradients, evaluations)


def _join_frames(*args: pd.DataFrame) -> pd.DataFrame:
    frames = [frame for frame in args if not frame.empty]
    return (
        frames[0].join(list(frames[1:]), how="outer") if len(frames) > 1 else frames[0]
    )


def _get_select(results: Results, field_name: str, sub_fields: set[str]) -> list[str]:
    results_field = getattr(results, field_name)
    return [
        name
        for name in {field.name for field in fields(results_field)}
        if f"{field_name}.{name}" in sub_fields
    ]


def _add_prefix(name: tuple[str, ...] | str, prefix: str) -> tuple[str, ...] | str:
    return (
        (f"{prefix}.{name[0]}",) + name[1:]
        if isinstance(name, tuple)
        else f"{prefix}.{name}"
    )


def _add_metadata(
    data_frame: pd.DataFrame, results: Results, sub_fields: set[str]
) -> pd.DataFrame:
    for field in sub_fields:
        split_fields = field.split(".")
        if split_fields[0] == "metadata":
            value = _get_value(results.metadata, split_fields[1:])
            if value is not None:
                data_frame[field] = value
    return data_frame


def _get_value(data: dict[str, Any], keys: list[str]) -> Any | None:  # noqa: ANN401
    for key in keys:
        if isinstance(data, dict):
            if key not in data:
                return None
            data = data[key]
        else:
            break
    return data



"""Plugin functionality for adding sampler plugins.

Sampler plugins are managed by a [`PluginManager`][ropt.plugins.PluginManager]
object, which returns classes or factory functions to create objects that
implement one or more sampling methods to produce perturbations. These objects
must adhere to the [`Sampler`][ropt.plugins.sampler.base.Sampler] abstract base
class.

Samplers can be added via the plugin manager, by default the
[`SciPySampler`][ropt.plugins.sampler.scipy.SciPySampler] plugin is installed
which provides a number of methods from the
[`scipy.stats`](https://docs.scipy.org/doc/scipy/reference/stats.html) package.
"""



import numpy as np
from numpy.typing import NDArray


def _get_failed_realizations(
    objectives: NDArray[np.float64],
    perturbed_objectives: NDArray[np.float64] | None,
    perturbation_min_success: int,
) -> NDArray[np.bool_]:
    failed_realizations = np.isnan(objectives[..., 0])
    if perturbed_objectives is not None:
        failed_pertubations = np.isnan(perturbed_objectives[..., 0])
        success_count = np.count_nonzero(~failed_pertubations, axis=-1)
        failed_realizations |= success_count < perturbation_min_success
    return np.array(failed_realizations)



"""This module implements the default meta data handler."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ropt.enums import EventType
from ropt.plugins.plan.base import ResultHandler

from ._utils import _get_set

if TYPE_CHECKING:
    from ropt.plan import Event, Plan


class DefaultMetadataHandler(ResultHandler):
    """The default metadata results handler.

    This handler adds arbitrary metadata to results produced by steps by merging
    them into the `metadata` field of the [`Results`][ropt.results.Results]
    objects that it receives. It is configured using a dictionary that maps keys
    to the data to store. These data entries can be of any type; however, if
    they the values is string starting with a single `$` it is assumed to be the
    key to value stored in the plan that should be inserted into the metadata.
    """

    def __init__(
        self, plan: Plan, *, data: dict[str, Any], tags: str | set[str] | None = None
    ) -> None:
        """Initialize a default metadata results handler.

        The data to merge into the metadata of a result are required.

        The `tags` field allows optional labels to be attached to each result,
        which can assist result handlers in filtering relevant results.

        Args:
            plan: The plan that runs this step.
            data: Data to merge into the metadata of the results.
            tags: Optional tags specifying which result sources to modify.
        """
        super().__init__(plan)

        self._data = data
        self._tags = _get_set(tags)

    def handle_event(self, event: Event) -> Event:
        """Handle an event.

        Args:
            event: The event to handle.

        Returns:
            The (possibly modified) event.
        """
        if (
            event.event_type
            in {
                EventType.FINISHED_EVALUATION,
                EventType.FINISHED_EVALUATOR_STEP,
            }
            and "results" in event.data
            and (event.tags & self._tags)
        ):
            for results in event.data["results"]:
                for key, value in self._data.items():
                    results.metadata[key] = (
                        self.plan[value[1:]]
                        if (
                            isinstance(value, str)
                            and value.startswith("$")
                            and not value[1:].startswith("$")
                        )
                        else value
                    )
        return event



"""Configuration class for objective functions."""

from __future__ import annotations

from typing import Self

import numpy as np
from pydantic import ConfigDict, model_validator

from ropt.config.utils import ImmutableBaseModel, normalize
from ropt.config.validated_types import (  # noqa: TC001
    Array1D,
    Array1DInt,
)


class ObjectiveFunctionsConfig(ImmutableBaseModel):
    """The configuration class for objective functions.

    This configuration class defines objective functions configured by the
    `objectives` field in an [`EnOptConfig`][ropt.config.enopt.EnOptConfig]
    object.

    `ropt` supports optimization over multiple objectives, which are summed after
    weighting with values passed via the `weights` field. This field is a
    `numpy` array, with a length that determines the number of objective functions.
    Its values will be normalized to have a sum equal to 1. For example, when
    `weights` is set to `[1, 1]`, the stored values will be `[0.5, 0.5]`.

    The objective functions may be subject to realization filters and function
    estimators. The `realization_filters` and `function_estimators` fields
    contain indices to the realization filter or function estimator objects to
    use. The objects referred to are configured in the parent
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    Attributes:
        weights:             Objective functions weights (default: 1.0).
        realization_filters: Optional realization filter indices.
        function_estimators: Optional function estimator indices.
    """

    weights: Array1D = np.array(1.0)
    realization_filters: Array1DInt | None = None
    function_estimators: Array1DInt | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _broadcast_and_normalize(self) -> Self:
        self._mutable()
        self.weights = normalize(self.weights)
        self._immutable()
        return self



"""Utilities for checking configuration values.

These utilities are intended to be used in the model validation code of Pydantic models.
"""

from collections.abc import Sequence as AbstractSequence
from collections.abc import Set as AbstractSet
from enum import IntEnum
from typing import Any, Sequence, Type, TypeVar, cast

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pydantic import BaseModel


def normalize(array: NDArray[np.float64]) -> NDArray[np.float64]:
    """Normalize a vector.

    Normalize the sum of the values to one.

    Args:
        array: The input array.

    Returns:
        ValueError: The normalized array
    """
    if array.sum() < np.finfo(np.float64).eps:
        msg = "the sum of weights is not positive"
        raise ValueError(msg)
    return immutable_array(array / array.sum())


def immutable_array(
    array_like: ArrayLike,
    **kwargs: Any,  # noqa: ANN401
) -> NDArray[Any]:
    """Make an immutable array.

    Converts the input to an array and makes it immutable.`

    Args:
        array_like: The input.
        kwargs    : Additional keyword arguments for array conversion.

    Returns:
        The immutable array.
    """
    array = np.array(array_like, **kwargs)
    array.setflags(write=False)
    return array


def broadcast_arrays(*args: Any) -> tuple[NDArray[Any], ...]:  # noqa: ANN401
    """Broadcast a set of arrays to a common dimensionality and makes them immutable.

    Args:
        args: The input arrays.

    Returns:
        The broadcasted immutable arrays.
    """
    results = np.broadcast_arrays(*args)
    return tuple(immutable_array(result) for result in results)


def broadcast_1d_array(array: NDArray[Any], name: str, size: int) -> NDArray[Any]:
    """Broadcast the input array to an 1D array of given size.

    Args:
        array: The input array.
        name:  The name of the array, used in an error message.
        size:  The size of the result.

    Returns:
        An 1D array of the requested size.
    """
    if size == 0:
        return immutable_array([], dtype=array.dtype)
    try:
        return np.broadcast_to(immutable_array(array), (size,))
    except ValueError as err:
        msg = f"{name} cannot be broadcasted to a length of {size}"
        raise ValueError(msg) from err


def check_enum_values(value: NDArray[np.ubyte], enum_type: Type[IntEnum]) -> None:
    """Check if an enum value is valid.

    Given an array of byte integers, check of the values are within the range of
    values of the given enum.

    Args:
        value:     The enum values.
        enum_type: The type to check.

    Raises:
        ValueError: If the array contains an invalid value.
    """
    min_enum = min(item.value for item in enum_type)
    max_enum = max(item.value for item in enum_type)
    if np.any(value < min_enum) or np.any(value > max_enum):
        msg = "invalid enumeration value"
        raise ValueError(msg)


def _convert_1d_array(array: ArrayLike | None) -> NDArray[np.float64] | None:
    if array is None:
        return array
    return immutable_array(array, dtype=np.float64, ndmin=1)


def _convert_1d_array_intc(array: ArrayLike | None) -> NDArray[np.intc] | None:
    if array is None:
        return array
    return immutable_array(array, dtype=np.intc, ndmin=1)


def _convert_1d_array_bool(
    array: ArrayLike | None,
) -> NDArray[np.bool_] | None:
    if array is None:
        return array
    return immutable_array(array, dtype=np.bool_, ndmin=1)


def _convert_2d_array(array: ArrayLike | None) -> NDArray[np.float64] | None:
    if array is None:
        return array
    return immutable_array(array, dtype=np.float64, ndmin=2)


def _convert_enum_array(array: ArrayLike | None) -> NDArray[np.ubyte] | None:
    if array is None:
        return array
    return immutable_array(array, dtype=np.ubyte, ndmin=1)


T = TypeVar("T")


def _convert_set(value: T | set[T] | Sequence[T]) -> set[T]:
    if isinstance(value, str):
        return {cast(T, value)}
    return set(value) if isinstance(value, AbstractSequence | AbstractSet) else {value}


def _convert_tuple(value: T | Sequence[T]) -> tuple[T, ...]:
    if isinstance(value, str):
        return (cast(T, value),)
    return tuple(value) if isinstance(value, AbstractSequence) else (value,)


class ImmutableBaseModel(BaseModel):
    """Base model for immutable classes.

    This model serves as an alternative to frozen Pydantic classes. It is
    particularly useful when post-initialization validators are required, as
    these validators may not function properly with frozen Pydantic classes.
    """

    _is_immutable: bool

    def _immutable(self) -> None:
        self._is_immutable = True

    def _mutable(self) -> None:
        self._is_immutable = False

    def __setattr__(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Attribute setter method.

        This method sets an attribute if the object is not immutable.

        Args:
            name:  The name of the attribute to set.
            value: The value to assign to the attribute.

        Raises:
            AttributeError: Raised if the object is immutable and cannot be modified.
        """
        if name != "_is_immutable" and self._is_immutable:
            msg = f"{self.__class__.__name__} is immutable"
            raise AttributeError(msg)
        super().__setattr__(name, value)



"""This module defines the abstract base classes for function estimators.

Function estimators can be added via the plugin mechanism to implement
additional ways to functions and gradient ensembles. Any object that adheres to
the
[`FunctionEstimator`][ropt.plugins.function_estimator.base.FunctionEstimator]
base class may be installed as a plugin.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ropt.plugins.base import Plugin

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


class FunctionEstimator(ABC):
    """Abstract base class for function estimators."""

    @abstractmethod
    def __init__(self, enopt_config: EnOptConfig, estimator_index: int) -> None:
        """Initialize the function estimator object.

        Args:
            enopt_config:    The configuration of the optimizer.
            estimator_index: The index of the estimator to use.
        """

    @abstractmethod
    def calculate_function(
        self, functions: NDArray[np.float64], weights: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Combine functions from realizations into an expected function.

        Tip: Calculation from merged realizations
           Normally the gradient is calculated for each realization separately
           and then combined into an overall gradient with `calculate_gradient`
           method. The `merge_realizations` flag in the ensemble optimizer
           configuration directs the optimizer to calculate the overall gradient
           from all realizations directly. This yields a reasonable estimation
           if the function estimator is an averaging operation, but may not be
           appropriate in other cases.

           At initialization, the `merge_realizations` flag should be checked,
           and if necessary a `ConfigError` with an appropriate message should
           be raised.

        Args:
            functions: The functions for each realization.
            weights:   The weight of each realization.

        Returns:
            The expected function values.
        """

    @abstractmethod
    def calculate_gradient(
        self,
        functions: NDArray[np.float64],
        gradient: NDArray[np.float64],
        weights: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Combine gradients from realizations into an expected gradient.

        Args:
            functions: The functions for each realization.
            gradient:  The gradient for each realization.
            weights:   The weight of each realization.

        Returns:
            The expected gradients.
        """


class FunctionEstimatorPlugin(Plugin):
    """The function estimator plugin base class."""

    @abstractmethod
    def create(
        self, enopt_config: EnOptConfig, estimator_index: int
    ) -> FunctionEstimator:
        """Initialize the function estimator object.

        Args:
            enopt_config:    The configuration of the optimizer.
            estimator_index: The index of the estimator to use.
        """



"""Plugin functionality for adding realization filters.

This package contains the abstract base class for realization filter plugins,
and the default realization filters that are part of `ropt`.

Realization filters are used by the optimizer to determine how a set of
realizations should be used to calculate objective and constraint function
values. They do this by calculating the weights that should be used for each
realization when calculating the values of a given set of objectives and
constraints.
"""
